#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""S3 storage utilities."""

import logging
import os
from typing import List, Optional, Set

from minio import Minio

from .exceptions import S3SyncError

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GT4SDS3Client:
    def __init__(
        self, host: str, access_key: str, secret_key: str, secure: bool = True
    ) -> None:
        """
        Construct an S3 client.

        Args:
            host: s3 host address.
            access_key: s3 access key.
            secret_key: s3 secret key.
            secure: whether the connection is secure or not. Defaults
                to True.
        """
        self.host = host
        self.access_key = access_key
        self.secret_key = secret_key
        self.secure = secure
        self.client = Minio(
            self.host,
            access_key=self.access_key,
            secret_key=self.secret_key,
            secure=self.secure,
        )

    def list_bucket_names(self) -> List[str]:
        """
        List all available s3 bucket names.

        Returns:
             List[str]: list with bucket names.
        """
        return [bucket.name for bucket in self.client.list_buckets()]

    def list_object_names(self, bucket: str, prefix: Optional[str] = None) -> List[str]:
        """
        List all available objects (recursive) in the given bucket based on a given prefix.

        Args:
            bucket: bucket name to search for objects.
            prefix: prefix for objects in the bucket.
                Defaults to None, a.k.a., no prefix filter.

        Returns:
            List[str]: list with object names.
        """
        return [
            s3_object.object_name
            for s3_object in self.client.list_objects(
                bucket_name=bucket, prefix=prefix, recursive=True
            )
        ]

    def list_directories(self, bucket: str, prefix: Optional[str] = None) -> Set[str]:
        """
        List all available "directories" in the given bucket based on a given prefix.

        Args:
            bucket: bucket name to search for objects.
            prefix: prefix for objects in the bucket.
                Defaults to None, a.k.a., no prefix filter.
                Needs to be a "directory" itself.

        Returns:
            List[str]: list with directory names.
        """
        if prefix:
            prefix = prefix + "/" if prefix[-1] != "/" else prefix
        return set(
            s3_object.object_name[len(prefix) if prefix else 0 : -1]
            for s3_object in self.client.list_objects(
                bucket_name=bucket, prefix=prefix, recursive=False
            )
            if s3_object.object_name[-1] == "/"
        )

    def upload_file(
        self, bucket: str, target_filepath: str, source_filepath: str
    ) -> None:
        """Upload a local file to S3 bucket.

        Args:
            bucket: bucket name to upload to.
            target_filepath: path to the file in S3.
            source_filepath: path to the file to upload.
        """
        self.client.fput_object(bucket, target_filepath, source_filepath)

    def sync_folder(
        self, bucket: str, path: str, prefix: Optional[str] = None, force: bool = False
    ) -> None:
        """Sync an entire folder from S3 recursively and save it under the given path.

        If :obj:`prefix` is given, every file under ``prefix/`` in S3 will be saver under ``path/`` in disk (i.e.
        ``prefix/`` is replaced by ``path/``).


        Args:
            bucket: bucket name to search for objects.
            path: path to save the objects in disk.
            prefix: prefix for objects in the bucket. Defaults to None, a.k.a., no prefix filter.
            force: force download even if a file with the same name is present. Defaults to False.
        """
        if not os.path.exists(path):
            logger.warning(f"path {path} does not exist, creating it...")
            os.makedirs(path)
        s3_objects = self.client.list_objects(
            bucket_name=bucket, prefix=prefix, recursive=True
        )
        for s3_object in s3_objects:
            object_name = s3_object.object_name
            object_name_stripped_prefix = (
                os.path.relpath(object_name, prefix) if prefix else object_name
            )
            filepath = os.path.join(path, object_name_stripped_prefix)
            # check for existence
            do_download = not os.path.exists(filepath)
            if do_download or force:
                logger.info(f"downloading file {object_name} in {filepath}")
                self.client.fget_object(
                    bucket_name=bucket, object_name=object_name, file_path=filepath
                )


def upload_file_to_s3(
    host: str,
    access_key: str,
    secret_key: str,
    bucket: str,
    target_filepath: str,
    source_filepath: str,
    secure: bool = True,
) -> None:
    """
    Sync the cache with the S3 remote storage.

    Args:
        host: s3 host address.
        access_key: s3 access key.
        secret_key: s3 secret key.
        bucket: bucket name to search for objects.
        target_filepath: path to save the objects in s3.
        source_filepath: path to the file to sync.
        secure: whether the connection is secure or not. Defaults
            to True.

    Raises:
        S3SyncError: in case of S3 syncing errors.
    """
    try:
        client = GT4SDS3Client(
            host=host, access_key=access_key, secret_key=secret_key, secure=secure
        )
        logger.info("starting syncing")
        client.upload_file(bucket, target_filepath, source_filepath)
        logger.info("syncing complete")
    except Exception:
        logger.exception("generic syncing error")
        raise S3SyncError(
            "UploadArtifactsErrors",
            f"error in uploading path={target_filepath} with host={host} and bucket={bucket}",
        )


def sync_folder_with_s3(
    host: str,
    access_key: str,
    secret_key: str,
    bucket: str,
    folder_path: str,
    prefix: Optional[str] = None,
    secure: bool = True,
) -> None:
    """
    Sync the cache with the S3 remote storage.

    Args:
        host: s3 host address.
        access_key: s3 access key.
        secret_key: s3 secret key.
        bucket: bucket name to search for objects.
        folder_path: folder path.
        prefix: prefix for objects in the bucket. Defaults to None, a.k.a., no prefix filter.
        secure: whether the connection is secure or not. Defaults
            to True.

    Raises:
        S3SyncError: in case of S3 syncing errors.
    """
    path = os.path.join(folder_path, prefix) if prefix else folder_path
    try:
        client = GT4SDS3Client(
            host=host, access_key=access_key, secret_key=secret_key, secure=secure
        )
        logger.info("starting syncing")
        client.sync_folder(bucket=bucket, path=path, prefix=prefix)
        logger.info("syncing complete")
    except Exception:
        logger.exception("generic syncing error")
        raise S3SyncError(
            "CacheSyncingError",
            f"error in syncing path={path} with host={host} access_key={access_key} secret_key={secret_key} secure={secure} bucket={bucket}",
        )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Losses for granular models."""

from typing import Any, Dict

import torch
from torch import nn


class MSLELossNegMix9(nn.Module):
    """MSLE loss negative mix 9."""

    def __init__(self) -> None:
        """Initialize the loss."""
        super().__init__()
        self.mse = nn.MSELoss(reduction="sum")

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> Any:
        """Forward pass in the loss.

        Args:
            prediction: predictions.
            target: groundtruth.

        Returns:
            loss value.
        """
        pred2 = prediction.clone()
        true2 = target.clone()
        pred2[pred2 < 0] = 0
        pred2 = pred2 + 1e-6
        true2 = true2 + 1e-6
        pred3 = prediction.clone()
        true3 = target.clone()
        pred3[target < 0.0001] = 0
        true3[target < 0.0001] = 0
        pred4 = prediction.clone()
        true4 = target.clone()
        pred4[pred4 > 0] = 0
        true4[true4 < 2] = 0
        l4_ = self.mse(pred4, true4)
        l1_ = self.mse(pred3 / (0.001 + true3), true3 / (0.001 + true3))
        l2_ = self.mse(prediction, target)
        l3_ = self.mse(torch.log(pred2), torch.log(true2))
        l0_ = torch.abs(l1_ * 0.1 + l2_ * 1.0 + l3_ * 1.0e-5 + l4_ * 10.0)
        return l0_


class MSLELossNegMix91(nn.Module):
    """MSLE loss negative mix 91."""

    def __init__(self):
        """Initialize the loss."""
        super().__init__()
        self.mse = nn.MSELoss(reduction="sum")
        self.mae = nn.L1Loss(reduction="sum")

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> Any:
        """Forward pass in the loss.

        Args:
            prediction: predictions.
            target: groundtruth.

        Returns:
            loss value.
        """
        pred2 = prediction.clone()
        true2 = target.clone()
        pred2[pred2 < 0] = 0
        pred2 = pred2 + 1e-6
        true2 = true2 + 1e-6
        l1_ = self.mae(prediction, target)
        l2_ = self.mse(prediction, target)
        l3_ = self.mse(torch.log(pred2), torch.log(true2))
        l0_ = torch.abs(l1_ * 0.3 + l2_ * 1.0 + l3_ * 1.0e-5)
        return l0_


class MseWithNans(nn.Module):
    """MSE with NaNs handling."""

    def __init__(self):
        """Initialize the loss."""
        super().__init__()

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> Any:
        """Forward pass in the loss.

        Args:
            prediction: predictions.
            target: groundtruth.

        Returns:
            loss value.
        """
        mask = torch.isnan(target)
        out = (prediction[~mask] - target[~mask]) ** 2
        loss = out.mean()
        return loss


class MaeWithNans(nn.Module):
    """MAE with NaNs handling."""

    def __init__(self):
        """Initialize the loss."""
        super().__init__()

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> Any:
        """Forward pass in the loss.

        Args:
            prediction: predictions.
            target: groundtruth.

        Returns:
            loss value.
        """
        mask = torch.isnan(target)
        if sum(mask) == len(target):
            return torch.tensor(0).type_as(prediction)
        out = abs((prediction[~mask] - target[~mask]))
        loss = sum(out) / len(prediction[~mask])
        return loss


class MSLELossNegMix92(nn.Module):
    """MSLE loss negative mix 92."""

    def __init__(self):
        """Initialize the loss."""
        super().__init__()
        self.mae = nn.L1Loss(reduction="mean")

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> Any:
        """Forward pass in the loss.

        Args:
            prediction: predictions.
            target: groundtruth.

        Returns:
            loss value.
        """
        l1_ = self.mae(prediction, target)
        mask = target.ge(0.001)
        l2_ = self.mae(prediction[mask], target[mask])
        l0_ = torch.abs(l1_ + l2_ * 10.0)
        return l0_


class MSLELossNegMix99(nn.Module):
    """MSLE loss negative mix 99."""

    def __init__(self):
        """Initialize the loss."""
        super().__init__()
        self.mse = nn.MSELoss(reduction="sum")

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> Any:
        """Forward pass in the loss.

        Args:
            prediction: predictions.
            target: groundtruth.

        Returns:
            loss value.
        """
        mask = torch.isnan(target)
        out = (prediction[~mask] - target[~mask]) ** 2
        loss = out.mean()
        return loss


LOSS_FACTORY: Dict[str, nn.Module] = {
    "mse": nn.MSELoss(),
    "mse-sum": nn.MSELoss(reduction="sum"),
    "mse-mean": nn.MSELoss(reduction="mean"),
    "mse-with-nans": MseWithNans(),
    "msewithnans": MseWithNans(),
    "mae": nn.L1Loss(),
    "mae-sum": nn.L1Loss(reduction="sum"),
    "mae-mean": nn.L1Loss(reduction="mean"),
    "mae-with-nans": MaeWithNans(),
    "maewithnans": MaeWithNans(),
    "bce": nn.BCELoss(),
    "bce-with-logits": nn.BCEWithLogitsLoss(),
    "bcewl": nn.BCEWithLogitsLoss(),
    "loss9": MSLELossNegMix9(),
    "l9": MSLELossNegMix9(),
    "msle-neg-mix-9": MSLELossNegMix9(),
    "loss91": MSLELossNegMix91(),
    "l91": MSLELossNegMix91(),
    "msle-neg-mix-91": MSLELossNegMix91(),
    "loss92": MSLELossNegMix92(),
    "l92": MSLELossNegMix92(),
    "msle-neg-mix-92": MSLELossNegMix92(),
    "loss99": MSLELossNegMix99(),
    "l99": MSLELossNegMix99(),
    "msle-neg-mix-99": MSLELossNegMix99(),
    "crossentropyloss": nn.CrossEntropyLoss(),
    "ce": nn.CrossEntropyLoss(),
}



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Model module."""

import ast
import copy
import logging
import os
from typing import Any, Dict, List, Optional, Tuple

import sentencepiece as _sentencepiece
import torch as _torch
import tensorflow as _tensorflow
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch_geometric.data as gd

from ..dataloader.dataset import GFlowNetDataset, GFlowNetTask
from ..envs.graph_building_env import GraphBuildingEnv, GraphBuildingEnvContext

# imports that have to be loaded before lightning to avoid segfaults
_sentencepiece
_tensorflow
_torch

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GFlowNetAlgorithm:
    """A generic algorithm for gflownet."""

    def compute_batch_losses(
        self, model: nn.Module, batch: gd.Batch, num_bootstrap: Optional[int] = 0
    ) -> Tuple[float, Dict[str, float]]:
        """Computes the loss for a batch of data, and proves logging informations.

        Args:
            model: the model being trained or evaluated.
            batch: a batch of graphs.
            num_bootstrap: the number of trajectories with reward targets in the batch (if applicable).

        Returns:
            loss: the loss for that batch.
            info: logged information about model predictions.
        """
        raise NotImplementedError()


class GFlowNetModule(pl.LightningModule):
    """Module from gflownet."""

    def __init__(
        self,
        configuration: Dict[str, Any],
        dataset: GFlowNetDataset,
        environment: GraphBuildingEnv,
        context: GraphBuildingEnvContext,
        task: GFlowNetTask,
        algorithm: GFlowNetAlgorithm,
        model: nn.Module,
    ) -> None:
        """Construct GFNModule.

        Args:
            configuration: the configuration of the module.
            dataset: the dataset to use.
            environment: the environment to use.
            context: the context to use.
            task: the task to solve.
            algorithm: algorithm (trajectory_balance or td_loss).
            model: architecture (graph_transformer_gfn or graph_transformer).
        """

        super().__init__()
        self.hps = configuration
        # self.save_hyperparameters()

        self.env = environment
        self.ctx = context
        self.dataset = dataset

        self.model = model
        self.algo = algorithm

        self.task = task

        self.test_output_path = self.hps["test_output_path"]

        self.rng = self.hps["rng"]
        self.mb_size = self.hps["global_batch_size"]
        self.clip_grad_param = self.hps["clip_grad_param"]
        self.sampling_tau = self.hps["sampling_tau"]

    def training_step(  # type:ignore
        self,
        batch: gd.Batch,
        batch_idx: int,
        optimizer_idx: int,
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Training step implementation.

        Args:
            batch: batch representation.
            epoch_idx: epoch index.
            batch_idx: batch index.

        Returns:
            loss and logs.
        """
        logs = dict()
        loss, info = self.algo.compute_batch_losses(
            self.model, batch, num_bootstrap=self.mb_size
        )
        logs.update(
            {
                self.model.name + f"/{k}": v.detach() if hasattr(v, "item") else v  # type: ignore
                for k, v in info.items()
            }
        )
        logs.update({"total_loss": loss.item()})  # type: ignore

        # logs for step
        _logs = {f"train/{k}": v for k, v in logs.items()}
        self.log_dict(_logs, on_step=True, on_epoch=True, prog_bar=True)
        self.log("train_loss", loss)

        # logs per epoch
        logs_epoch = {f"train_epoch/{k}": v for k, v in logs.items()}
        logs_epoch["step"] = self.current_epoch
        self.log_dict(
            logs_epoch,
            on_step=False,
            on_epoch=True,
            prog_bar=False,
        )

        return {"loss": loss, "logs": logs}

    def training_step_end(self, batch_parts):
        for i in self.model.parameters():
            self.clip_grad_callback(i)

        if self.sampling_tau > 0:
            for a, b in zip(self.model.parameters(), self.sampling_model.parameters()):
                b.data.mul_(self.sampling_tau).add_(a.data * (1 - self.sampling_tau))

    def validation_step(  # type:ignore
        self, batch: gd.Batch, batch_idx: int, *args: Any, **kwargs: Any
    ) -> Dict[str, Any]:
        """Validation step implementation.

        Args:
            batch: batch representation.

        Returns:
            loss and logs.
        """
        loss = 0.0
        logs = dict()
        loss, info = self.algo.compute_batch_losses(
            self.model, batch, num_bootstrap=batch.num_offline
        )
        logs.update({k: v if hasattr(v, "item") else v for k, v in info.items()})
        logs.update({"total_loss": loss})

        self.log_dict(
            {f"val/{k}": v for k, v in logs.items()},
            on_step=True,
            on_epoch=True,
            prog_bar=False,
        )
        self.log("val_loss", loss)
        return {"loss": loss, "logs": logs}

    def test_step(  # type:ignore
        self, batch: Any, batch_idx: int
    ) -> Dict[str, Any]:
        """Testing step implementation.

        Args:
            batch: batch representation.
            batch_idx: batch index, unused.

        Returns:
            loss, logs, and latent encodings.
        """
        loss = 0.0
        logs = dict()
        loss, info = self.algo.compute_batch_losses(
            self.model, batch, num_bootstrap=batch.num_offline
        )
        logs.update({k: v if hasattr(v, "item") else v for k, v in info.items()})
        logs.update({"total_loss": loss})

        self.log_dict(
            {f"test/{k}": v for k, v in logs.items()},
            on_step=True,
            on_epoch=True,
            prog_bar=True,
        )
        self.log("test_loss", loss)
        return {"loss": loss, "logs": logs}

    def prediction_step(self, batch) -> torch.Tensor:
        """Inference step.

        Args:
            batch: batch data.

        Returns:
            output forward.
        """
        return self(batch)

    def train_epoch_end(self, outputs: List[Dict[str, Any]]):
        """Train epoch end.

        Args:
            outputs: list of outputs epoch.

        Returns:
        """
        pass

    # change the following to new implementation
    def test_epoch_end(self, outputs: List[Dict[str, Any]]) -> None:  # type:ignore
        """Callback called at the end of an epoch on test outputs.

        Dump encodings and targets for the test set.

        Args:
            outputs: outputs for test batches.
        """
        z = {}
        targets = {}
        z_keys = [key for key in outputs[0]["z"]]
        targets_keys = [key for key in outputs[0]["targets"]]

        for key in z_keys:
            z[key] = (
                torch.cat(
                    [torch.squeeze(an_output["z"][key]) for an_output in outputs], dim=0
                )
                .detach()
                .cpu()
                .numpy()
            )

        for key in targets_keys:
            targets[key] = (
                torch.cat(
                    [torch.squeeze(an_output["targets"][key]) for an_output in outputs],
                    dim=0,
                )
                .detach()
                .cpu()
                .numpy()
            )

        pd.to_pickle(z, f"{self.test_output_path}{os.path.sep}z_build.pkl")
        pd.to_pickle(targets, f"{self.test_output_path}{os.path.sep}targets.pkl")

    def configure_optimizers(self):
        """Configure optimizers.

        Returns:
            an optimizer, currently only Adam is supported.
        """
        # Separate z parameters from non-z to allow for LR decay on the former
        z_params = list(self.model.log_z.parameters())  # type: ignore
        non_z_params = [
            i for i in self.model.parameters() if all(id(i) != id(j) for j in z_params)
        ]

        self.opt = torch.optim.Adam(
            non_z_params,
            self.hps["learning_rate"],
            (self.hps["momentum"], 0.999),
            weight_decay=self.hps["weight_decay"],
            eps=self.hps["adam_eps"],
        )
        self.opt_z = torch.optim.Adam(z_params, self.hps["learning_rate"], (0.9, 0.999))

        self.lr_sched = torch.optim.lr_scheduler.LambdaLR(
            self.opt, lambda steps: 2 ** (-steps / self.hps["lr_decay"])
        )
        self.lr_sched_z = torch.optim.lr_scheduler.LambdaLR(
            self.opt_z, lambda steps: 2 ** (-steps / self.hps["z_lr_decay"])
        )

        if self.sampling_tau > 0:
            self.sampling_model = copy.deepcopy(self.model)
        else:
            self.sampling_model = self.model

        eps = self.hps["tb_epsilon"]
        self.hps["tb_epsilon"] = ast.literal_eval(eps) if isinstance(eps, str) else eps

        self.clip_grad_callback = {
            "value": (
                lambda params: torch.nn.utils.clip_grad_value_(
                    params, self.clip_grad_param
                )
            ),
            "norm": (
                lambda params: torch.nn.utils.clip_grad_norm_(
                    params, self.clip_grad_param
                )
            ),
            "none": (lambda x: None),
        }[self.hps["clip_grad_type"]]

        return [self.opt, self.opt_z], [self.lr_sched, self.lr_sched_z]



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Algortihms for keyword generation using BERT models."""

import logging
from dataclasses import field
from typing import Any, Callable, ClassVar, Dict, Iterable, Optional, Set, TypeVar

from ...core import (
    AlgorithmConfiguration,
    GeneratorAlgorithm,
    get_configuration_class_with_attributes,
)
from ...registry import ApplicationsRegistry
from .implementation import KeyBERT

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=Any)
Targeted = Callable[[T], Iterable[Any]]


class KeywordBERTGenerationAlgorithm(GeneratorAlgorithm[S, T]):
    """Topics prediction algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ):
        """Instantiate KeywordBERTGenerationAlgorithm ready to predict topics.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for predicting topics for a given text::

                config = KeyBERTGenerator()
                algorithm = KeywordBERTGenerationAlgorithm(configuration=config, target="This is a text I want to understand better")
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Targeted[T]:
        """Get the function to perform the prediction via KeywordBERTGenerationAlgorithm's generator.

        Args:
            configuration: helps to set up specific application of KeywordBERTGenerationAlgorithm.
            target: context or condition for the generation.

        Returns:
            callable with target generating keywords sorted by relevance.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Any = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.predict


@ApplicationsRegistry.register_algorithm_application(KeywordBERTGenerationAlgorithm)
class KeyBERTGenerator(AlgorithmConfiguration[str, str]):
    """Configuration to generate keywords.

    If the  model is not found in the cache, models are collected from https://www.sbert.net/docs/pretrained_models.html.
    distilbert-base-nli-stsb-mean-tokens is recommended for english, while xlm-r-bert-base-nli-stsb-mean-tokens for all
    other languages as it support 100+ languages.
    """

    algorithm_name: ClassVar[str] = KeywordBERTGenerationAlgorithm.__name__
    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "nlp"
    algorithm_version: str = "distilbert-base-nli-mean-tokens"

    minimum_keyphrase_ngram: int = field(
        default=1,
        metadata=dict(description=("Lower bound for phrase size.")),
    )
    maximum_keyphrase_ngram: int = field(
        default=2,
        metadata=dict(description=("Upper bound for phrase size.")),
    )
    stop_words: str = field(
        default="english",
        metadata=dict(description=("Language for the stop words removal.")),
    )
    top_n: int = field(
        default=10,
        metadata=dict(description=("Number of keywords to extract.")),
    )
    use_maxsum: bool = field(
        default=False,
        metadata=dict(
            description=("Control usage of max sum similarity for keywords generated.")
        ),
    )
    use_mmr: bool = field(
        default=False,
        metadata=dict(
            description=(
                "Control usage of max marginal relevance for keywords generated."
            )
        ),
    )
    diversity: float = field(
        default=0.5,
        metadata=dict(description=("Diversity for the results when enabling use_mmr.")),
    )
    number_of_candidates: int = field(
        default=20,
        metadata=dict(description=("Candidates considered when enabling use_maxsum.")),
    )

    def get_target_description(self) -> Dict[str, str]:
        """Get description of the target for generation.

        Returns:
            target description.
        """
        return {
            "title": "Text to analyze",
            "description": "Text considered for the keyword generation task.",
            "type": "string",
        }

    def get_conditional_generator(self, resources_path: str) -> KeyBERT:
        """Instantiate the actual generator implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`generate_batch<gt4sd.algorithms.conditional_generation.key_bert.implementation.KeyBERT.predict>` method for targeted generation.
        """
        return KeyBERT(
            resources_path=resources_path,
            minimum_keyphrase_ngram=self.minimum_keyphrase_ngram,
            maximum_keyphrase_ngram=self.maximum_keyphrase_ngram,
            stop_words=self.stop_words,
            top_n=self.top_n,
            use_maxsum=self.use_maxsum,
            use_mmr=self.use_mmr,
            diversity=self.diversity,
            number_of_candidates=self.number_of_candidates,
            model_name=self.algorithm_version,
        )

    @classmethod
    def list_versions(cls) -> Set[str]:
        """Get possible algorithm versions.

        Standard S3 and cache search adding the version used in the configuration.

        Returns:
            viable values as :attr:`algorithm_version` for the environment.
        """
        logger.warning(
            "more algorithm versions can be found on https://www.sbert.net/docs/pretrained_models.html"
        )
        return (
            get_configuration_class_with_attributes(cls)
            .list_versions()
            .union({cls.algorithm_version})
        )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
from enum import Enum
from typing import List, Optional, Union

import importlib_resources
import numpy as np
import pandas as pd
import torch
import yaml
from gt4sd_molformer.finetune.finetune_pubchem_light import (
    LightningModule as RegressionLightningModule,
)
from gt4sd_molformer.finetune.finetune_pubchem_light import (
    PropertyPredictionDataModule as RegressionDataModule,
)
from gt4sd_molformer.finetune.finetune_pubchem_light import (
    PropertyPredictionDataset as RegressionDataset,
)
from gt4sd_molformer.finetune.finetune_pubchem_light_classification import (
    LightningModule as ClassificationLightningModule,
)
from gt4sd_molformer.finetune.finetune_pubchem_light_classification import (
    PropertyPredictionDataModule as ClassificationDataModule,
)
from gt4sd_molformer.finetune.finetune_pubchem_light_classification import (
    PropertyPredictionDataset as ClassificationDataset,
)
from gt4sd_molformer.finetune.finetune_pubchem_light_classification_multitask import (
    MultitaskEmbeddingDataset,
    MultitaskModel,
    PropertyPredictionDataModule,
)
from gt4sd_molformer.finetune.ft_tokenizer.ft_tokenizer import MolTranBertTokenizer
from paccmann_generator.drug_evaluators import SIDER
from paccmann_generator.drug_evaluators import ClinTox as _ClinTox
from paccmann_generator.drug_evaluators import OrganDB as _OrganTox
from paccmann_generator.drug_evaluators import SCScore
from paccmann_generator.drug_evaluators import Tox21 as _Tox21
from pydantic import ConfigDict, Field
from tdc import Oracle
from tdc.metadata import download_receptor_oracle_name

from ...algorithms.core import (
    ConfigurablePropertyAlgorithmConfiguration,
    Predictor,
    PredictorAlgorithm,
)
from ...domains.materials import SmallMolecule
from ...frameworks.torch import device_claim
from ..core import (
    ApiTokenParameters,
    CallablePropertyPredictor,
    ConfigurableCallablePropertyPredictor,
    DomainSubmodule,
    IpAdressParameters,
    PropertyPredictorParameters,
    PropertyValue,
    S3Parameters,
)
from ..utils import (
    docking_import_check,
    get_activity_fn,
    get_similarity_fn,
    to_smiles,
    validate_api_token,
    validate_ip,
)
from .functions import (
    bertz,
    esol,
    is_scaffold,
    lipinski,
    logp,
    molecular_weight,
    number_of_aromatic_rings,
    number_of_atoms,
    number_of_h_acceptors,
    number_of_h_donors,
    number_of_heterocycles,
    number_of_large_rings,
    number_of_rings,
    number_of_rotatable_bonds,
    number_of_stereocenters,
    plogp,
    qed,
    sas,
    tpsa,
)


# NOTE: property prediction parameters
class ScscoreConfiguration(PropertyPredictorParameters):
    score_scale: int = 5
    fp_len: int = 1024
    fp_rad: int = 2


class SimilaritySeedParameters(PropertyPredictorParameters):
    smiles: str = Field(..., examples=["c1ccccc1"])
    fp_key: str = "ECFP4"


class ActivityAgainstTargetParameters(PropertyPredictorParameters):
    target: str = Field(..., examples=["drd2"], description="name of the target.")


class AskcosParameters(IpAdressParameters):
    class Output(str, Enum):
        plausability: str = "plausibility"
        num_step: str = "num_step"
        synthesizability: str = "synthesizability"
        price: str = "price"

    output: Output = Field(
        default=Output.plausability,
        examples=[Output.synthesizability],
        description="Main output return type from ASKCOS",
    )
    save_json: bool = Field(default=False)
    file_name: str = Field(default="tree_builder_result.json")
    num_trials: int = Field(default=5)
    max_depth: int = Field(default=9)
    max_branching: int = Field(default=25)
    expansion_time: int = Field(default=60)
    max_ppg: int = Field(default=100)
    template_count: int = Field(default=1000)
    max_cum_prob: float = Field(default=0.999)
    chemical_property_logic: str = Field(default="none")
    max_chemprop_c: int = Field(default=0)
    max_chemprop_n: int = Field(default=0)
    max_chemprop_o: int = Field(default=0)
    max_chemprop_h: int = Field(default=0)
    chemical_popularity_logic: str = Field(default="none")
    min_chempop_reactants: int = Field(default=5)
    min_chempop_products: int = Field(default=5)
    filter_threshold: float = Field(default=0.1)
    return_first: str = Field(default="true")
    model_config = ConfigDict(use_enum_values=True)


class MoleculeOneParameters(ApiTokenParameters):

    oracle_name: str = "Molecule One Synthesis"


class DockingTdcParameters(PropertyPredictorParameters):
    # To dock against a receptor defined via TDC
    target: str = Field(
        ...,
        examples=download_receptor_oracle_name,
        description="Target for docking, provided via TDC",
    )


class DockingParameters(PropertyPredictorParameters):
    # To dock against a user-provided receptor
    name: str = Field(default="pyscreener")
    receptor_pdb_file: str = Field(
        examples=["/tmp/2hbs.pdb"], description="Path to receptor PDB file"
    )
    box_center: List[int] = Field(
        examples=[[15.190, 53.903, 16.917]], description="Docking box center"
    )
    box_size: List[float] = Field(
        examples=[[20, 20, 20]], description="Docking box size"
    )


class S3ParametersMolecules(S3Parameters):
    domain: DomainSubmodule = DomainSubmodule("molecules")


class MolformerParameters(S3ParametersMolecules):
    algorithm_name: str = "molformer"
    batch_size: int = Field(description="Prediction batch size", default=128)
    workers: int = Field(description="Number of data loading workers", default=8)
    device: Optional[str] = Field(
        description="Device to be used for inference", default=None
    )


class MolformerClassificationParameters(MolformerParameters):
    algorithm_application: str = "classification"


class MolformerMultitaskClassificationParameters(MolformerParameters):
    algorithm_application: str = "multitask_classification"


class MolformerRegressionParameters(MolformerParameters):
    algorithm_application: str = "regression"


class MCAParameters(S3ParametersMolecules):
    algorithm_name: str = "MCA"


class Tox21Parameters(MCAParameters):
    algorithm_application: str = "Tox21"


class ClinToxParameters(MCAParameters):
    algorithm_application: str = "ClinTox"


class SiderParameters(MCAParameters):
    algorithm_application: str = "SIDER"


class OrganToxParameters(MCAParameters):
    class Organs(str, Enum):
        adrenal_gland: str = "Adrenal Gland"
        bone_marrow: str = "Bone Marrow"
        brain: str = "Brain"
        eye: str = "Eye"
        heart: str = "Heart"
        kidney: str = "Kidney"
        liver: str = "Liver"
        lung: str = "Lung"
        lymph_node: str = "Lymph Node"
        mammary_gland: str = "Mammary Gland"
        ovary: str = "Ovary"
        pancreas: str = "Pancreas"
        pituitary_gland: str = "Pituitary Gland"
        spleen: str = "Spleen"
        stomach: str = "Stomach"
        testes: str = "Testes"
        thymus: str = "Thymus"
        thyroid_gland: str = "Thyroid Gland"
        urinary_bladder: str = "Urinary Bladder"
        uterus: str = "Uterus"

    class ToxType(str, Enum):
        chronic: str = "chronic"
        subchronic: str = "subchronic"
        multigenerational: str = "multigenerational"
        all: str = "all"

    algorithm_application: str = "OrganTox"
    site: Organs = Field(
        ...,
        examples=[Organs.kidney],
        description="name of the target site of interest.",
    )
    toxicity_type: ToxType = Field(
        default=ToxType.all,
        examples=[ToxType.chronic],
        description="type of toxicity for which predictions are made.",
    )


class _Molformer(PredictorAlgorithm):
    """Base class for all Molformer predictive algorithms."""

    def __init__(self, parameters: MolformerParameters):

        # Set up the configuration from the parameters
        configuration = ConfigurablePropertyAlgorithmConfiguration(
            algorithm_type=parameters.algorithm_type,
            domain=parameters.domain,
            algorithm_name=parameters.algorithm_name,
            algorithm_application=parameters.algorithm_application,
            algorithm_version=parameters.algorithm_version,
        )

        self.batch_size = parameters.batch_size
        self.workers = parameters.workers

        self.tokenizer_path = (
            importlib_resources.files("gt4sd_molformer") / "finetune/bert_vocab.txt"
        )

        self.device = device_claim(parameters.device)

        # The parent constructor calls `self.get_model`.
        super().__init__(configuration=configuration)

    def get_resources_path_and_config(self, resources_path: str):

        model_path = os.path.join(resources_path, "model.ckpt")
        config_path = os.path.join(resources_path, "hparams.yaml")

        with open(config_path, "r") as stream:
            config = yaml.safe_load(stream)

        return config, model_path


class MolformerClassification(_Molformer):
    """Class for all Molformer classification algorithms."""

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """

        config, model_path = self.get_resources_path_and_config(resources_path)

        config["num_workers"] = 0

        tokenizer = MolTranBertTokenizer(self.tokenizer_path)

        model = ClassificationLightningModule(config, tokenizer).load_from_checkpoint(
            model_path,
            strict=False,
            config=config,
            tokenizer=tokenizer,
            vocab=len(tokenizer.vocab),
        )

        model.to(self.device)
        model.eval()

        # Wrapper to get the predictions
        def informative_model(samples: Union[str, List[str]]) -> List[float]:

            if isinstance(samples, str):
                samples = [samples]

            df = pd.DataFrame.from_dict({"smiles": samples})

            dataset = ClassificationDataset(df)
            datamodule = ClassificationDataModule(config, tokenizer)
            datamodule.test_ds = dataset

            preds = []
            for batch in datamodule.test_dataloader():
                with torch.no_grad():
                    batch = [x.to(self.device) for x in batch]
                    batch_output = model.testing_step(batch, 0, 0)

                preds_cpu = batch_output["pred"][:, 1]
                y_pred = np.where(preds_cpu >= 0.5, 1, 0)
                preds += y_pred.tolist()

            return preds

        return informative_model


class MolformerMultitaskClassification(_Molformer):
    """Class for all Molformer multitask classification algorithms."""

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """

        config, model_path = self.get_resources_path_and_config(resources_path)

        config["num_workers"] = 0

        tokenizer = MolTranBertTokenizer(self.tokenizer_path)

        model = MultitaskModel(config, tokenizer).load_from_checkpoint(
            model_path,
            strict=False,
            config=config,
            tokenizer=tokenizer,
            vocab=len(tokenizer.vocab),
        )

        model.to(self.device)
        model.eval()

        # Wrapper to get the predictions
        def informative_model(samples: Union[str, List[str]]) -> List[str]:

            if isinstance(samples, str):
                samples = [samples]

            df = pd.DataFrame.from_dict({"smiles": samples})

            dataset = MultitaskEmbeddingDataset(df)
            datamodule = PropertyPredictionDataModule(config, tokenizer)
            datamodule.test_ds = dataset

            preds = []
            for batch in datamodule.test_dataloader():

                with torch.no_grad():
                    batch = [x.to(self.device) for x in batch]
                    batch_output = model.testing_step(batch, 0, 0)

                batch_preds_idx = torch.argmax(batch_output["pred"], dim=1)
                batch_preds = [config["measure_names"][i] for i in batch_preds_idx]
                preds += batch_preds

            return preds

        return informative_model


class MolformerRegression(_Molformer):
    """Class for all Molformer regression algorithms."""

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """

        config, model_path = self.get_resources_path_and_config(resources_path)

        config["num_workers"] = 0

        tokenizer = MolTranBertTokenizer(self.tokenizer_path)

        model = RegressionLightningModule(config, tokenizer).load_from_checkpoint(
            model_path,
            strict=False,
            config=config,
            tokenizer=tokenizer,
            vocab=len(tokenizer.vocab),
        )

        model.to(self.device)
        model.eval()

        # Wrapper to get the predictions
        def informative_model(samples: Union[str, List[str]]) -> List[float]:

            if isinstance(samples, str):
                samples = [samples]

            df = pd.DataFrame.from_dict({"smiles": samples})

            dataset = RegressionDataset(df, False, config["aug"])
            datamodule = RegressionDataModule(config, tokenizer)
            datamodule.test_ds = dataset

            preds = []
            for batch in datamodule.test_dataloader():
                with torch.no_grad():
                    batch = [x.to(self.device) for x in batch]
                    batch_output = model.testing_step(batch, 0, 0)

                preds += batch_output["pred"].view(-1).tolist()

            return preds

        return informative_model


# NOTE: property prediction classes
class Plogp(CallablePropertyPredictor):
    """Calculate the penalized logP of a molecule. This is the logP minus the number of
    rings with > 6 atoms minus the SAS.
    """

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=plogp, parameters=parameters)


class Lipinski(CallablePropertyPredictor):
    """Calculate whether a molecule adheres to the Lipinski-rule-of-5.
    A crude approximation of druglikeness.
    """

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=lipinski, parameters=parameters)


class Esol(CallablePropertyPredictor):
    """Estimate the water solubility of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=esol, parameters=parameters)


class Scscore(CallablePropertyPredictor):
    """Calculate the synthetic complexity score (SCScore) of a molecule."""

    def __init__(
        self, parameters: ScscoreConfiguration = ScscoreConfiguration()
    ) -> None:
        super().__init__(
            callable_fn=SCScore(**parameters.dict()), parameters=parameters
        )


class Sas(CallablePropertyPredictor):
    """Calculate the synthetic accessibility score (SAS) for a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=sas, parameters=parameters)


class Bertz(CallablePropertyPredictor):
    """Calculate Bertz index of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=bertz, parameters=parameters)


class Tpsa(CallablePropertyPredictor):
    """Calculate the total polar surface area of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=tpsa, parameters=parameters)


class Logp(CallablePropertyPredictor):
    """Calculates the partition coefficient of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=logp, parameters=parameters)


class Qed(CallablePropertyPredictor):
    """Calculate the quantitative estimate of drug-likeness (QED) of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=qed, parameters=parameters)


class NumberHAcceptors(CallablePropertyPredictor):
    """Calculate number of H acceptors of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_h_acceptors, parameters=parameters)


class NumberAtoms(CallablePropertyPredictor):
    """Calculate number of atoms of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_atoms, parameters=parameters)


class NumberHDonors(CallablePropertyPredictor):
    """Calculate number of H donors of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_h_donors, parameters=parameters)


class NumberAromaticRings(CallablePropertyPredictor):
    """Calculate number of aromatic rings of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_aromatic_rings, parameters=parameters)


class NumberRings(CallablePropertyPredictor):
    """Calculate number of rings of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_rings, parameters=parameters)


class NumberRotatableBonds(CallablePropertyPredictor):
    """Calculate number of rotatable bonds of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_rotatable_bonds, parameters=parameters)


class NumberLargeRings(CallablePropertyPredictor):
    """Calculate the amount of large rings (> 6 atoms) of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_large_rings, parameters=parameters)


class MolecularWeight(CallablePropertyPredictor):
    """Calculate molecular weight of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=molecular_weight, parameters=parameters)


class IsScaffold(CallablePropertyPredictor):
    """Whether a molecule is identical to its Murcko Scaffold."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=is_scaffold, parameters=parameters)


class NumberHeterocycles(CallablePropertyPredictor):
    """The amount of heterocycles of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_heterocycles, parameters=parameters)


class NumberStereocenters(CallablePropertyPredictor):
    """The amount of stereo centers of a molecule."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=number_of_stereocenters, parameters=parameters)


class SimilaritySeed(CallablePropertyPredictor):
    """Calculate the similarity of a molecule to a seed molecule."""

    def __init__(self, parameters: SimilaritySeedParameters) -> None:
        super().__init__(
            callable_fn=get_similarity_fn(
                target_mol=parameters.smiles, fp_key=parameters.fp_key
            ),
            parameters=parameters,
        )


class ActivityAgainstTarget(CallablePropertyPredictor):
    """Calculate the activity of a molecule against a target molecule."""

    def __init__(self, parameters: ActivityAgainstTargetParameters) -> None:
        super().__init__(
            callable_fn=get_activity_fn(target=parameters.target), parameters=parameters
        )


class Askcos(ConfigurableCallablePropertyPredictor):
    """
    A property predictor that uses the ASKCOs API to calculate the synthesizability
    of a molecule.
    """

    def __init__(self, parameters: AskcosParameters):

        # Raises if IP is not valid
        msg = (
            "You have to point to an IP address of a running ASKCOS instance. "
            "For details on setting this up, see: https://tdcommons.ai/functions/oracles/#askcos"
        )
        if not isinstance(parameters.host_ip, str):
            raise TypeError(f"IP adress must be a string, not {parameters.host_ip}")

        if not hasattr(parameters, "host_ip"):
            raise AttributeError(f"IP adress missing in {parameters}")

        if "http" not in parameters.host_ip:
            raise ValueError(
                f"ASKCOS requires an IP prepended with a http, e.g., "
                f"'http://xx.xx.xxx.xxx' and not {parameters.host_ip}."
            )
        ip = parameters.host_ip.split("//")[1]

        validate_ip(ip, message=msg)
        super().__init__(callable_fn=Oracle(name="ASKCOS"), parameters=parameters)


class MoleculeOne(CallablePropertyPredictor):
    """
    A property predictor that uses the MoleculeOne API to calculate the synthesizability
    of a molecule.
    """

    def __init__(self, parameters: MoleculeOneParameters):

        msg = (
            "You have to provide a valid API key, for details on setting this up, see: "
            "https://tdcommons.ai/functions/oracles/#moleculeone"
        )

        # Only performs type checking on API key
        validate_api_token(parameters, message=msg)

        super().__init__(
            callable_fn=Oracle(
                name=parameters.oracle_name, api_token=parameters.api_token
            ),
            parameters=parameters,
        )


class DockingTdc(ConfigurableCallablePropertyPredictor):
    """
    A property predictor that computes the docking score against a target
    provided via the TDC package (see: https://tdcommons.ai/functions/oracles/#docking-scores)
    """

    def __init__(self, parameters: DockingTdcParameters):

        docking_import_check()
        callable = Oracle(name=parameters.target)
        super().__init__(callable_fn=callable, parameters=parameters)


class Docking(ConfigurableCallablePropertyPredictor):
    """
    A property predictor that computes the docking score against a user-defined target.
    Relies on TDC backend, see https://tdcommons.ai/functions/oracles/#docking-scores for setup.
    """

    def __init__(self, parameters: DockingParameters):

        docking_import_check()
        callable = Oracle(
            name=parameters.name,
            receptor_pdb_file=parameters.receptor_pdb_file,
            box_center=parameters.box_center,
            box_size=parameters.box_size,
        )
        super().__init__(callable_fn=callable, parameters=parameters)


class _MCA(PredictorAlgorithm):
    """Base class for all MCA-based predictive algorithms."""

    def __init__(self, parameters: MCAParameters):

        # Set up the configuration from the parameters
        configuration = ConfigurablePropertyAlgorithmConfiguration(
            algorithm_type=parameters.algorithm_type,
            domain=parameters.domain,
            algorithm_name=parameters.algorithm_name,
            algorithm_application=parameters.algorithm_application,
            algorithm_version=parameters.algorithm_version,
        )

        # The parent constructor calls `self.get_model`.
        super().__init__(configuration=configuration)


class Tox21(_MCA):
    """Model to predict environmental toxicity for the 12 endpoints in Tox21."""

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """
        # This model returns a singular reward and not a prediction for all 12 classes.
        model = _Tox21(model_path=resources_path)

        # Wrapper to get toxicity-endpoint-level predictions
        def informative_model(x: SmallMolecule) -> List[PropertyValue]:
            x = to_smiles(x)
            _ = model(x)
            return model.predictions.detach().tolist()

        return informative_model

    @classmethod
    def get_description(cls) -> str:
        text = """
        This model predicts the 12 endpoints from the Tox21 challenge.
        The endpoints are: NR-AR, NR-AR-LBD, NR-AhR, NR-Aromatase, NR-ER, NR-ER-LBD, NR-PPAR-gamma, SR-ARE, SR-ATAD5, SR-HSE, SR-MMP, SR-p53
        For details on the data see: https://tripod.nih.gov/tox21/challenge/.
        """
        return text


class ClinTox(_MCA):
    """Model to predict environmental toxicity for the 12 endpoints in Tox21."""

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """
        # This model returns a singular reward and not a prediction for both classes.
        model = _ClinTox(model_path=resources_path)

        # Wrapper to get toxicity-endpoint-level predictions
        def informative_model(x: SmallMolecule) -> List[PropertyValue]:
            x = to_smiles(x)
            _ = model(x)
            return model.predictions.detach().tolist()

        return informative_model

    @classmethod
    def get_description(cls) -> str:
        text = """
        This model is a multitask classifier for two classes:
            1. Predicted probability to receive FDA approval.
            2. Predicted probability of failure in clinical trials.
        For details on the data see: https://pubs.rsc.org/en/content/articlehtml/2018/sc/c7sc02664a.
        """
        return text


class Sider(_MCA):
    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """
        # This model returns a singular reward and not a prediction for both classes.
        model = SIDER(model_path=resources_path)

        # Wrapper to get toxicity-endpoint-level predictions
        def informative_model(x: SmallMolecule) -> List[PropertyValue]:
            x = to_smiles(x)
            _ = model(x)
            return model.predictions.detach().tolist()

        return informative_model

    @classmethod
    def get_description(cls) -> str:
        text = """
        This model is a multitask classifier to predict side effects of drugs across
        27 classes. For details on the data see:
        https://pubs.rsc.org/en/content/articlehtml/2018/sc/c7sc02664a.
        """
        return text


class OrganTox(_MCA):
    """Model to predict toxicity for different organs."""

    def __init__(self, parameters: OrganToxParameters) -> None:

        # Extract model-specific parameters
        self.site = parameters.site
        self.toxicity_type = parameters.toxicity_type

        super().__init__(parameters=parameters)

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """
        # This model returns a singular reward and not a prediction for both classes.
        model = _OrganTox(
            model_path=resources_path, site=self.site, toxicity_type=self.toxicity_type
        )

        # Wrapper to get toxicity-endpoint-level predictions
        def informative_model(x: SmallMolecule) -> List[PropertyValue]:
            x = to_smiles(x)
            _ = model(x)
            all_preds = model.predictions.detach()

            return all_preds[model.class_indices].tolist()

        return informative_model

    @classmethod
    def get_description(cls) -> str:
        text = """
        This model is a multitask classifier to toxicity across different organs and
        toxicity types (`chronic`, `subchronic`, `multigenerational` or `all`).
        The organ has to specified in the constructor. The toxicity type defaults to
        `all` in which case three values are returned in the order `chronic`,
        `multigenerational` and `subchronic`.

        For details on the data see:
        https://pubs.rsc.org/en/content/articlehtml/2018/sc/c7sc02664a.
        """
        return text



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Model utilities."""


class KLAnnealer:
    """Annealer scaling KL weights (beta) linearly according to the number of epochs."""

    def __init__(
        self, kl_low: float, kl_high: float, n_epochs: int, start_epoch: int
    ) -> None:
        """Construct KLAnnealer.

        Args:
            kl_low: low KL weight.
            kl_high: high KL weight.
            n_epochs: number of epochs.
            start_epoch: starting epoch.
        """
        self.kl_low = kl_low
        self.kl_high = kl_high
        self.n_epochs = n_epochs
        self.start_epoch = start_epoch
        self.kl = (self.kl_high - self.kl_low) / (self.n_epochs - self.start_epoch)

    def __call__(self, epoch: int) -> float:
        """Call the annealer.

        Args:
            epoch: current epoch number.

        Returns:
            the beta weight.
        """
        k = (epoch - self.start_epoch) if epoch >= self.start_epoch else 0
        beta = self.kl_low + k * self.kl
        if beta > self.kl_high:
            beta = self.kl_high
        return beta



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from enum import Enum
from typing import Any, Callable, Union

from pydantic import BaseModel, Field

PropertyValue = Union[float, int]


class DomainSubmodule(str, Enum):
    molecules: str = "molecules"
    properties: str = "properties"
    crystals: str = "crystals"


class PropertyPredictorParameters(BaseModel):
    """Abstract class for property computation."""

    pass


# Base class for property predictors that use S3 artifacts
class S3Parameters(PropertyPredictorParameters):

    algorithm_type: str = "prediction"

    domain: DomainSubmodule = Field(
        ..., examples=["molecules"], description="Submodule of gt4sd.properties"
    )
    algorithm_name: str = Field(
        ..., examples=["MCA"], description="Name of the algorithm"
    )
    algorithm_version: str = Field(
        ..., examples=["v0"], description="Version of the algorithm"
    )
    algorithm_application: str = Field(..., examples=["Tox21"])


class ApiTokenParameters(PropertyPredictorParameters):
    api_token: str = Field(
        ...,
        examples=["apk-c9db......"],
        description="The API token/key to access the service",
    )


class IpAdressParameters(PropertyPredictorParameters):

    host_ip: str = Field(
        ...,
        examples=["xx.xx.xxx.xxx"],
        description="The host IP address to access the service",
    )


class PropertyPredictor:
    """PropertyPredictor base class."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        """Construct a PropertyPredictor using the related parameters.
        Args:
            parameters: parameters to configure the predictor.
        """
        self.parameters = parameters

    def __call__(self, sample: Any) -> PropertyValue:
        """Call the PropertyPredictor.

        Args:
            sample: a sample to use for predicting the property of interest.

        Returns:
            Property:

        Example:
            An example for predicting properties::

                property_predictor = PropertyPredictor(parameters)
                value = property_predictor(sample)
        """
        raise NotImplementedError


class CallablePropertyPredictor(PropertyPredictor):
    """Property predictor based on a callable."""

    def __init__(
        self,
        callable_fn: Callable,
        parameters: PropertyPredictorParameters = PropertyPredictorParameters(),
    ) -> None:
        self.callable_fn = callable_fn
        super().__init__(parameters=parameters)

    def __call__(self, sample: Any) -> PropertyValue:
        """Call the PropertyPredictor.

        Args:
            sample: a sample to use for predicting the property of interest.

        Returns:
            Property: Property predicted by the predictor.

        Example:
            An example for predicting properties::

                property_predictor = CallablePropertyPredictor(callable_fn=lambda a: id(a), parameters)
                value = property_predictor(sample)
        """
        return self.callable_fn(sample)


class ConfigurableCallablePropertyPredictor(CallablePropertyPredictor):
    """Property predictor based on a callable that is configured using the provided parameters."""

    def __call__(self, sample: Any) -> PropertyValue:
        """Call the PropertyPredictor.

        Args:
            sample: a sample to use for predicting the property of interest.

        Returns:
            Property:

        Example:
            An example for predicting properties::

                property_predictor = CallablePropertyPredictor(callable_fn=lambda a, b: id(a), parameters)
                value = property_predictor(sample)
        """
        return self.callable_fn(sample, **self.parameters.dict())



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation of PaccMann^RL conditional generators."""

import logging
from typing import List

import torch
from rdkit import Chem

from ...conditional_generation.paccmann_rl.core import (
    PaccMannRL,
    PaccMannRLProteinBasedGenerator,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class PaccMannVaeDefaultGenerator:
    """
    Molecular generator as implemented in https://doi.org/10.1016/j.isci.2021.102269
    """

    def __init__(
        self,
        temperature: float = 1.4,
        batch_size: int = 32,
        algorithm_version: str = "v0",
        generated_length: int = 100,
    ) -> None:
        """
        Initialize the generator.

        Args:
            batch_size: batch size used for generation.
            algorithm_version: algorithm version for the PaccMannRLProteinBasedGenerator.
                NOTE: Only the decoder of that model is used here.
            temperature: temperature for the sampling. Defaults to 1.4.
            generated_length: maximum length of the generated molecules.
                Defaults to 100.
        """
        self.configuration = PaccMannRLProteinBasedGenerator(
            algorithm_version=algorithm_version,
            temperature=temperature,  # type: ignore
            generated_length=generated_length,  # type: ignore
            batch_size=batch_size,  # type: ignore
        )
        self.batch_size = batch_size

        self.algorithm = PaccMannRL(configuration=self.configuration, target="")
        self.model = self.configuration.get_conditional_generator(
            self.algorithm.local_artifacts
        )

    def generate(self) -> List[str]:
        """
        Generate a given number of samples (molecules) from a given protein.

        Args:
            number_of_molecules: number of molecules to sample.

        Returns:
            list of SMILES generated.
        """
        smiles: List = []
        while len(smiles) < self.batch_size:
            # Define latent code
            latent = torch.randn(1, self.batch_size, self.model.encoder_latent_size)
            # Bypass algorithm.sample by decoding SMILES directly from latent
            generated_smiles = self.model.get_smiles_from_latent(latent)
            _, valid_ids = self.model.validate_molecules(generated_smiles)
            valid_ids = [
                i
                for i in valid_ids
                if len(
                    Chem.DetectChemistryProblems(
                        Chem.MolFromSmiles(generated_smiles[i])
                    )
                )
                == 0
            ]
            generated_molecules = list([generated_smiles[index] for index in valid_ids])
            smiles.extend(generated_molecules)
        return smiles



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Implementation details for PGT algorithms.
"""

import logging
import os
import re
from typing import List, Optional, Tuple, Union

import torch
from transformers import GPT2LMHeadModel, GPT2Tokenizer

from ....frameworks.torch import device_claim

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

MAXIMUM_LENGTH = int(10000)
GENERATION_PROMPTS = {
    "title-to-abstract": "{} <|sep|> Given the above title, suggest an abstract <|sep|>",
    "abstract-to-claim": "{} <|sep|> Given the above abstract, suggest a claim <|sep|>",
    "claim-to-abstract": "{} <|sep|> Given the above claim, suggest an abstract <|sep|>",
    "abstract-to-title": "{} <|sep|> Given the above abstract, suggest a title <|sep|>",
}
EDITING_TYPES = ["abstract", "claim"]
COHERENCE_TYPES = ["title-abstract", "abstract-claim", "title-claim"]

STOPPING_PUNCTUATION_REGEX = re.compile(r"(.+(?=\.|!|;)(.|!|;)|(.*))")


def adjust_length_to_model(length: int, maximum_sequence_length: int):
    """Adjust sequence length.
    Args:
        length: target length.
        maximum_sequence_length: maximum sequence length.
    Returns:
        the adjusted length.
    """
    if length < 0 and maximum_sequence_length > 0:
        logger.warning(
            f"negative length, adjusting to model supported length {maximum_sequence_length}"
        )
        length = maximum_sequence_length
    elif 0 < maximum_sequence_length < length:
        logger.warning(
            f"longer then model supported length, adjusting to {maximum_sequence_length}"
        )
        length = maximum_sequence_length
    elif length < 0:
        logger.warning(f"negative length, adjusting to maximal length {MAXIMUM_LENGTH}")
        length = MAXIMUM_LENGTH
    return length


class Generator:
    """Implementation of a generator."""

    def __init__(
        self,
        resources_path: str,
        model_type: str,
        model_name: str,
        max_length: int,
        top_k: int,
        top_p: float,
        num_return_sequences: int,
        prompt: str = "This is an interesting prompt",
        no_repeat_ngram_size: int = 2,
        device: Optional[Union[torch.device, str]] = None,
    ):
        """PGT generation algorithm.
        Args:
            resources_path: path to the cache.
            model_type: type of the model.
            model_name: name of the model weights/version.
            prompt: prompt for text generation.
            max_length: max length of the generated text.
            top_k: number of top-k probability token to keep.
            top_p: only tokens with cumulative probabilities summing up to this value are kept.
            num_return_sequences: number of generated sequences.
            no_repeat_ngram_size: size of n-gram to not appear twice.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        self.device = device_claim(device)
        self.resources_path = resources_path
        self.model_type = model_type
        self.model_name = model_name
        self.prompt = prompt
        self.length = max_length
        self.k = top_k
        self.p = top_p
        self.no_repeat_ngram_size = no_repeat_ngram_size
        self.number_of_sequences = num_return_sequences
        self.load_model()

    def load_model(self) -> None:
        """Load a pretrained PGT model."""

        if (
            os.path.exists(self.resources_path)
            and len(os.listdir(self.resources_path)) > 0
        ):
            model_name_or_path = self.resources_path
        else:
            logger.error(f"{self.resources_path} not found")

        self.tokenizer = GPT2Tokenizer.from_pretrained(  # type:ignore
            model_name_or_path,
            sep_token="<|sep|>",
            mask_token="[MASK]",
            pad_token="<|pad|>",
            additional_special_tokens=["<|mask_sep|>"],
        )
        self.model = GPT2LMHeadModel.from_pretrained(model_name_or_path)
        self.model.resize_token_embeddings(len(self.tokenizer))
        self.model.to(self.device)
        # adjusting length
        self.length = adjust_length_to_model(
            self.length, self.model.config.max_position_embeddings
        )

    def generate_case(self) -> Union[List[str], List[Tuple[str, ...]]]:
        """Sample text snippets.

        Returns:
            generated text snippets.
        """

        self.prompt = re.sub(" +", " ", self.prompt)
        self.prompt = re.sub(r'\s([?.!,:;·"](?:\s|$))', r"\1", self.prompt)

        encoded_prompt = self.tokenizer.encode(self.prompt, return_tensors="pt")
        encoded_prompt = encoded_prompt.to(self.device)

        if encoded_prompt.size()[-1] == 0:
            input_ids = None
        else:
            input_ids = encoded_prompt

        output_sequences = self.model.generate(
            input_ids=input_ids,
            max_length=self.length,
            top_k=self.k,
            top_p=self.p,
            do_sample=True,
            no_repeat_ngram_size=self.no_repeat_ngram_size,
            num_return_sequences=self.number_of_sequences,
        )

        # NOTE: remove the batch dimension when returning multiple sequences
        if len(output_sequences.shape) > 2:
            output_sequences.squeeze_()

        generated_sequences: List[str] = []

        for generated_sequence in output_sequences:
            generated_sequence = generated_sequence.tolist()
            text = self.tokenizer.decode(generated_sequence)
            text = text.replace(self.prompt, "")
            text = text.split("<|endoftext|>")[0]
            text = text.replace("<|pad|>", "")
            text = text.strip()
            text = STOPPING_PUNCTUATION_REGEX.search(  # type:ignore
                text
            ).group()

            generated_sequences.append(text)

        return self.format_output(self.prompt, generated_sequences)

    def format_output(
        self, input_text: Union[str, Tuple[str]], generated_sequences: List[str]
    ) -> Union[List[str], List[Tuple[str, ...]]]:
        """Format output. In the general case just return the generated sequences.

        Args:
            input_text: generation input.
            generated_sequences: generated sequences.

        Returns:
            formatted generated sequences.
        """
        return generated_sequences


class PartGenerator(Generator):
    """Implementation of edit generator."""

    def __init__(
        self,
        resources_path: str,
        input_text: str,
        model_type: str,
        model_name: str,
        task: str,
        max_length: int,
        top_k: int,
        top_p: float,
        num_return_sequences: int,
        no_repeat_ngram_size: int = 2,
        device: Optional[Union[torch.device, str]] = None,
    ):
        """PGT generation algorithm.
        Args:
            resources_path: path to the cache.
            input_text: input text for generation.
            task: generation task.
            model_type: type of the model.
            model_name: name of the model weights/version.
            max_length: max length of the generated text.
            top_k: number of top-k probability token to keep.
            top_p: only tokens with cumulative probabilities summing up to this value are kept.
            num_return_sequences: number of generated sequences.
            no_repeat_ngram_size: size of n-gram to not appear twice.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """

        if task not in GENERATION_PROMPTS:
            raise ValueError(f"{task} is not a valid option for task.")

        prompt = GENERATION_PROMPTS[task]
        prompt = prompt.format(input_text)

        super().__init__(
            resources_path=resources_path,
            model_type=model_type,
            model_name=model_name,
            prompt=prompt,
            max_length=max_length,
            top_k=top_k,
            top_p=top_p,
            num_return_sequences=num_return_sequences,
            no_repeat_ngram_size=no_repeat_ngram_size,
            device=device,
        )


class EditGenerator(Generator):
    """Implementation of edit generator."""

    def __init__(
        self,
        resources_path: str,
        input_text: str,
        model_type: str,
        model_name: str,
        max_length: int,
        top_k: int,
        top_p: float,
        num_return_sequences: int,
        no_repeat_ngram_size: int = 2,
        device: Optional[Union[torch.device, str]] = None,
        input_type: str = "abstract",
    ):
        """PGT generation algorithm.
        Args:
            resources_path: path to the cache.
            input_text: input text for generation.
            model_type: type of the model.
            model_name: name of the model weights/version.
            max_length: max length of the generated text.
            top_k: number of top-k probability token to keep.
            top_p: only tokens with cumulative probabilities summing up to this value are kept.
            num_return_sequences: number of generated sequences.
            no_repeat_ngram_size: size of n-gram to not appear twice.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
            input_type: part of a patent the input text belongs.
        """

        if input_type not in EDITING_TYPES:
            raise ValueError(
                f"{input_type} is not a valid option for editing input type."
            )

        prompt = f"{input_text} <|sep|> Replace the [MASK] tokens in the above {input_type} <|sep|>"

        super().__init__(
            resources_path=resources_path,
            model_type=model_type,
            model_name=model_name,
            prompt=prompt,
            max_length=max_length,
            top_k=top_k,
            top_p=top_p,
            num_return_sequences=num_return_sequences,
            no_repeat_ngram_size=no_repeat_ngram_size,
            device=device,
        )

    def format_output(
        self, input_text: Union[str, Tuple[str]], generated_sequences: List[str]
    ) -> Union[List[str], List[Tuple[str, ...]]]:
        """Format output for the patent editing task.

        Args:
           input_text: generation input.
           generated_sequences: generated sequences.

        Returns:
           formatted generated sequences.
        """

        filtered_generated_sequences = []

        number_of_masks = input_text.count("[MASK]") - 1

        for seq in generated_sequences:
            generated_masked_tokens = seq.split("<|mask_sep|>")

            filtered_generated_sequence = []
            for i in range(number_of_masks):
                if i < len(generated_masked_tokens):
                    gen_token = generated_masked_tokens[i]
                else:
                    gen_token = "[MASK NOT PREDICTED]"
                filtered_generated_sequence.append(gen_token.strip())

            filtered_generated_sequences.append(tuple(filtered_generated_sequence))

        return filtered_generated_sequences


class CoherenceCheckGenerator(Generator):
    """Implementation of coherence check generator."""

    def __init__(
        self,
        resources_path: str,
        input_a: str,
        input_b: str,
        model_type: str,
        model_name: str,
        max_length: int,
        top_k: int,
        top_p: float,
        num_return_sequences: int,
        no_repeat_ngram_size: int = 2,
        device: Optional[Union[torch.device, str]] = None,
        coherence_type: str = "title-abstract",
    ):
        """PGT generation algorithm.
        Args:
            resources_path: path to the cache.
            input_a: first input for coherence check.
            input_b: second input for coherence check.
            model_type: type of the model.
            model_name: name of the model weights/version.
            max_length: max length of the generated text.
            top_k: number of top-k probability token to keep.
            top_p: only tokens with cumulative probabilities summing up to this value are kept.
            num_return_sequences: number of generated sequences.
            no_repeat_ngram_size: size of n-gram to not appear twice.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
            coherence_type: input types for the check.
        """

        type_a, type_b = self.extract_coherence_types(coherence_type)

        prompt = f"{input_a} <|sep|> {input_b} <|sep|> Do the above {type_a} and {type_b} belong to the same patent? <|sep|>"

        super().__init__(
            resources_path=resources_path,
            model_type=model_type,
            model_name=model_name,
            prompt=prompt,
            max_length=max_length,
            top_k=top_k,
            top_p=top_p,
            num_return_sequences=num_return_sequences,
            no_repeat_ngram_size=no_repeat_ngram_size,
            device=device,
        )

    def extract_coherence_types(self, coherence_type: str) -> Tuple[str, str]:
        """Check the validity and extract coherence types of input text.

        Args:
            coherence_type: Input types of the coherence check.
        Returns:
            tuple containing the type of the input.
        """

        if coherence_type in COHERENCE_TYPES:
            type_a, type_b = coherence_type.split("-")

            return type_a, type_b

        else:
            raise ValueError(f"{coherence_type} is not a valid coherence type.")

    def format_output(
        self, input_text: Union[str, Tuple[str]], generated_sequences: List[str]
    ) -> Union[List[str], List[Tuple[str, ...]]]:
        """Format output for the patent coherence task.

        Args:
           input_text: generation input.
           generated_sequences: generated sequences.

        Returns:
           formatted generated sequences.
        """

        if (
            "yes" in generated_sequences[0].lower()
            and "no" not in generated_sequences[0].lower()
        ):
            return ["yes"]
        elif (
            "no" in generated_sequences[0].lower()
            and "yes" not in generated_sequences[0].lower()
        ):
            return ["no"]
        else:
            return ["NA"]



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import List, Tuple, cast

from rdkit import Chem


def get_substructure_indices(
    full_sequence: List[str], substructure: List[str]
) -> List[int]:
    """
    Args:
        full_sequence: A list of strings, each representing a token from the full sequence
        substructure: A list of strings, each representing a token from the substructure that
            is contained in the full sequence.

    Returns:
        A list of integers, corresponding to all the indices of the tokens in the full sequence
        that match the substructure.

    """
    substructure_indices: List = []
    for i in range(len(full_sequence)):
        if full_sequence[i] == substructure[0]:
            if full_sequence[i : i + len(substructure)] == substructure:
                substructure_indices.extend(range(i, i + len(substructure)))
    return substructure_indices


def filter_stubbed(
    property_sequences: Tuple[Tuple[str, str]], target: str, threshold: float = 0.5
) -> Tuple[Tuple[str, str]]:
    """
    Remove stub-like molecules that are substantially smaller than the target.

    Args:
        sequences: List of generated molecules.
        properties: Properties of the molecules. Only used to be returned after filtering.
        target: Seed molecule.
        threshold: Fraction of size of generated molecule compared to seed determining the
            threshold under which molecules are discarded. Defaults to 0.5.

    Returns:
        Tuple of tuples of length 2 with filtered, generated molecule and its properties.
    """

    seed = Chem.MolFromSmiles(target)  # type: ignore

    seed_atoms = len(list(seed.GetAtoms()))  # type: ignore
    seed_bonds = seed.GetNumBonds()  # type: ignore

    smis: List[str] = []
    props: List[str] = []
    for smi, prop in property_sequences:
        if smi == "":
            continue
        try:
            mol = Chem.MolFromSmiles(smi)  # type: ignore
            num_atoms = len(list(mol.GetAtoms()))  # type: ignore
            num_bonds = mol.GetNumBonds()  # type: ignore

            if num_atoms > (threshold * seed_atoms) and num_bonds > (
                threshold * seed_bonds
            ):
                smis.append(smi)
                props.append(prop)
        except Exception:
            continue

    successes = cast(Tuple[Tuple[str, str]], tuple(zip(smis, props)))
    return successes



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from pydantic import Field

from ..core import (
    CallablePropertyPredictor,
    ConfigurableCallablePropertyPredictor,
    PropertyPredictorParameters,
)
from .functions import (
    aliphatic_index,
    aromaticity,
    boman_index,
    charge,
    charge_density,
    hydrophobic_ratio,
    instability,
    isoelectric_point,
    length,
    molecular_weight,
)


# NOTE: property prediction parameters
class AmideConfiguration(PropertyPredictorParameters):
    amide: bool = Field(
        False,
        examples=[False],
        description="whether the sequences are C-terminally amidated.",
    )


class PhConfiguration(PropertyPredictorParameters):
    ph: float = 7.0


class AmidePhConfiguration(PropertyPredictorParameters):
    amide: bool = Field(
        False,
        examples=[False],
        description="whether the sequences are C-terminally amidated.",
    )
    ph: float = 7.0


# NOTE: property prediction classes
class Length(CallablePropertyPredictor):
    """Retrieves the number of residues of a protein."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=length, parameters=parameters)


class MolecularWeight(ConfigurableCallablePropertyPredictor):
    """Computes the molecular weight of a protein."""

    def __init__(self, parameters: AmideConfiguration) -> None:
        super().__init__(callable_fn=molecular_weight, parameters=parameters)


class BomanIndex(CallablePropertyPredictor):
    """Computes the Boman index of a protein (sum of solubility values of all residues)."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=boman_index, parameters=parameters)


class AliphaticIndex(CallablePropertyPredictor):
    """Computes the aliphatic index of a protein. Measure of thermal stability."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=aliphatic_index, parameters=parameters)


class HydrophobicRatio(CallablePropertyPredictor):
    """Computes the hydrophobicity of a protein, relative freq. of **A,C,F,I,L,M & V**."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=hydrophobic_ratio, parameters=parameters)


class Charge(ConfigurableCallablePropertyPredictor):
    """Computes the charge of a protein."""

    def __init__(self, parameters: AmidePhConfiguration) -> None:
        super().__init__(callable_fn=charge, parameters=parameters)


class ChargeDensity(ConfigurableCallablePropertyPredictor):
    """Computes the charge density of a protein."""

    def __init__(self, parameters: AmidePhConfiguration) -> None:
        super().__init__(callable_fn=charge_density, parameters=parameters)


class IsoelectricPoint(ConfigurableCallablePropertyPredictor):
    """Computes the isoelectric point of every residue and aggregates."""

    def __init__(self, parameters: AmideConfiguration) -> None:
        super().__init__(callable_fn=isoelectric_point, parameters=parameters)


class Aromaticity(CallablePropertyPredictor):
    """Computes aromaticity of the protein (relative frequency of Phe+Trp+Tyr)."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=aromaticity, parameters=parameters)


class Instability(CallablePropertyPredictor):
    """Calculates the protein instability."""

    def __init__(
        self, parameters: PropertyPredictorParameters = PropertyPredictorParameters()
    ) -> None:
        super().__init__(callable_fn=instability, parameters=parameters)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Model combiner module."""

import os
from typing import Any, Callable, Dict, List, Tuple, cast

import sentencepiece as _sentencepiece
import torch as _torch
import tensorflow as _tensorflow
import pandas as pd
import pytorch_lightning as pl
import torch

from .models import GranularBaseModel, GranularEncoderDecoderModel
from .models.model_builder import building_models, define_latent_models_input_size

# imports that have to be loaded before lightning to avoid segfaults
_sentencepiece
_tensorflow
_torch


class GranularModule(pl.LightningModule):
    """Module from granular."""

    def __init__(
        self,
        architecture_autoencoders: List[Dict[str, Any]],
        architecture_latent_models: List[Dict[str, Any]],
        lr: float = 1e-4,
        test_output_path: str = "./test",
        **kwargs,
    ) -> None:
        """Construct GranularModule.

        Args:
            architecture_autoencoders: list of autoencoder architecture configurations.
            architecture_latent_models: list of latent model architecture configurations.
            lr: learning rate for Adam optimizer. Defaults to 1e-4.
            test_output_path: path where to save latent encodings and predictions for the test set
                when an epoch ends. Defaults to a a folder called "test" in the current working directory.
        """
        super().__init__()
        self.save_hyperparameters()

        architecture_latent_models = define_latent_models_input_size(
            architecture_autoencoders, architecture_latent_models
        )
        self.architecture_autoencoders = architecture_autoencoders
        self.architecture_latent_models = architecture_latent_models

        self.autoencoders = building_models(self.architecture_autoencoders)
        self.latent_models = building_models(self.architecture_latent_models)

        self.lr = lr
        self.test_output_path = test_output_path
        for model in self.autoencoders + self.latent_models:
            setattr(self, model.name, model)

    def _autoencoder_step(
        self, batch: Any, model: GranularEncoderDecoderModel, model_step_fn: Callable
    ) -> Tuple[Any, Any, Any]:
        """Autoencoder module forward pass.

        Args:
            batch: batch representation.
            model: a module.
            model_step_fn: callable for the step.

        Returns:
            a tuple containing the latent representation, the loss and the logs for the module.
        """
        return model_step_fn(
            input_data=batch[model.input_key],
            target_data=batch[model.target_key],
            device=self.device,
            current_epoch=self.current_epoch,
        )

    def _latent_step(
        self,
        batch: Any,
        model: GranularBaseModel,
        model_step_fn: Callable,
        z: Dict[int, Any],
    ) -> Tuple[Any, Any, Any]:
        """Latent module forward pass.

        Args:
            batch: batch representation.
            model: a module.
            model_step_fn: callable for the step.
            z: latent encodings.

        Returns:
            a tuple containing the latent step ouput, the loss and the logs for the module.
        """
        z_model_input = torch.cat(
            [
                torch.squeeze(z[pos]) if len(z[pos].size()) == 3 else z[pos]
                for pos in model.from_position
            ],
            dim=1,
        )
        return model_step_fn(
            input_data=z_model_input,
            target_data=batch[model.target_key],
            device=self.device,
            current_epoch=self.current_epoch,
        )

    def training_step(  # type:ignore
        self, batch: Any, *args, **kwargs
    ) -> Dict[str, Any]:
        """Training step implementation.

        Args:
            batch: batch representation.

        Returns:
            loss and logs.
        """
        loss = 0.0
        z = dict()
        logs = dict()

        for model in self.autoencoders:
            z[model.position], loss_model, logs_model = self._autoencoder_step(
                batch=batch,
                model=cast(GranularEncoderDecoderModel, model),
                model_step_fn=model.step,
            )
            logs.update({model.name + f"/{k}": v for k, v in logs_model.items()})
            loss += loss_model

        for model in self.latent_models:
            _, loss_model, logs_model = self._latent_step(
                batch=batch, model=model, model_step_fn=model.step, z=z
            )
            logs.update({model.name + f"/{k}": v for k, v in logs_model.items()})
            loss += loss_model

        logs.update({"total_loss": loss})
        self.log_dict(
            {f"train/{k}": v for k, v in logs.items()}, on_epoch=False, prog_bar=False
        )
        logs_epoch = {f"train_epoch/{k}": v for k, v in logs.items()}
        logs_epoch["step"] = self.current_epoch
        self.log_dict(logs_epoch, on_step=False, on_epoch=True, prog_bar=False)

        return {"loss": loss, "logs": logs}

    def validation_step(  # type:ignore
        self, batch: Any, *args, **kwargs
    ) -> Dict[str, Any]:
        """Validation step implementation.

        Args:
            batch: batch representation.

        Returns:
            loss and logs.
        """
        loss = 0.0
        z = dict()
        logs = dict()

        for model in self.autoencoders:
            z[model.position], loss_model, logs_model = self._autoencoder_step(
                batch=batch,
                model=cast(GranularEncoderDecoderModel, model),
                model_step_fn=model.val_step,
            )
            logs.update({model.name + f"/{k}": v for k, v in logs_model.items()})
            loss += loss_model

        for model in self.latent_models:
            _, loss_model, logs_model = self._latent_step(
                batch=batch, model=model, model_step_fn=model.val_step, z=z
            )
            logs.update({model.name + f"/{k}": v for k, v in logs_model.items()})
            loss += loss_model

        logs.update({"total_loss": loss})
        self.log_dict(
            {f"val/{k}": v for k, v in logs.items()}, on_epoch=True, prog_bar=True
        )

        return {"loss": loss, "logs": logs}

    def test_step(  # type:ignore
        self, batch: Any, batch_idx: int, *args, **kwargs
    ) -> Dict[str, Any]:
        """Testing step implementation.

        Args:
            batch: batch representation.
            batch_idx: batch index, unused.

        Returns:
            loss, logs, and latent encodings.
        """
        loss = 0.0
        z = dict()
        logs = dict()

        for model in self.autoencoders:
            z[model.position], loss_model, logs_model = self._autoencoder_step(
                batch=batch,
                model=cast(GranularEncoderDecoderModel, model),
                model_step_fn=model.val_step,
            )
            logs.update({model.name + f"/{k}": v for k, v in logs_model.items()})
            loss += loss_model

        for model in self.latent_models:
            _, loss_model, logs_model = self._latent_step(
                batch=batch, model=model, model_step_fn=model.val_step, z=z
            )
            logs.update({model.name + f"/{k}": v for k, v in logs_model.items()})
            loss += loss_model

        logs.update({"total_loss": loss})
        self.log_dict(
            {f"val/{k}": v for k, v in logs.items()}, on_epoch=True, prog_bar=True
        )
        return {"loss": loss, "logs": logs, "z": z}

    def test_epoch_end(self, outputs: List[Dict[str, Any]]) -> None:  # type:ignore
        """Callback called at the end of an epoch on test outputs.

        Dump encodings and targets for the test set.

        Args:
            outputs: outputs for test batches.
        """
        z = {}
        targets = {}
        z_keys = [key for key in outputs[0]["z"]]
        targets_keys = [key for key in outputs[0]["targets"]]
        for key in z_keys:
            z[key] = (
                torch.cat(
                    [torch.squeeze(an_output["z"][key]) for an_output in outputs], dim=0
                )
                .detach()
                .cpu()
                .numpy()
            )

        for key in targets_keys:
            targets[key] = (
                torch.cat(
                    [torch.squeeze(an_output["targets"][key]) for an_output in outputs],
                    dim=0,
                )
                .detach()
                .cpu()
                .numpy()
            )

        pd.to_pickle(z, f"{self.test_output_path}{os.path.sep}z_build.pkl")
        pd.to_pickle(targets, f"{self.test_output_path}{os.path.sep}targets.pkl")

    def configure_optimizers(self) -> torch.optim.Optimizer:
        """Configure optimizers.

        Returns:
            an optimizer, currently only Adam is supported.
        """
        return torch.optim.Adam(self.parameters(), lr=self.lr)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Graph GA implementation."""

from guacamol_baselines.graph_ga.goal_directed_generation import GB_GA_Generator


class GraphGA:
    def __init__(
        self,
        smi_file,
        mutation_rate: float,
        population_size: int,
        offspring_size: int,
        n_jobs: int,
        random_start: bool,
        generations: int,
        patience: int,
    ):
        """Initialize SMILESGA.

        Args:
            smi_file: path where to load hypothesis, candidate labels and, optionally, the smiles file.
            population_size: used with n_mutations for the initial generation of smiles within the population.
            n_jobs: number of concurrently running jobs.
            random_start: set to True to randomly choose list of SMILES for generating optimizied molecules.
            generations: number of evolutionary generations.
            patience: used for early stopping if population scores remains the same after generating molecules.
            mutation_rate: frequency of the new mutations in a single gene or organism over time.
            offspring_size: number of molecules to select for new population.
        """
        self.smi_file = smi_file
        self.mutation_rate = mutation_rate
        self.population_size = population_size
        self.offspring_size = offspring_size
        self.n_jobs = n_jobs
        self.random_start = random_start
        self.generations = generations
        self.patience = patience

    def get_generator(self) -> GB_GA_Generator:
        """Create an instance of the GB_GA_Generator.

        Returns:
            an instance of GB_GA_Generator.
        """
        optimiser = GB_GA_Generator(
            smi_file=self.smi_file,
            population_size=self.population_size,
            offspring_size=self.offspring_size,
            mutation_rate=self.mutation_rate,
            generations=self.generations,
            n_jobs=self.n_jobs,
            random_start=self.random_start,
            patience=self.patience,
        )
        return optimiser



#
# MIT License
#
# Copyright (c) 2024 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Enzeptional processing tests."""
import pytest
import numpy as np
from gt4sd.frameworks.enzeptional import (
    HuggingFaceModelLoader,
    HuggingFaceTokenizerLoader,
    HuggingFaceEmbedder,
    sanitize_intervals,
    round_up,
    sanitize_intervals_with_padding,
    SelectionGenerator,
    CrossoverGenerator,
)


@pytest.fixture
def huggingface_embedder():
    model_loader = HuggingFaceModelLoader()
    tokenizer_loader = HuggingFaceTokenizerLoader()

    language_model_path = "facebook/esm2_t33_650M_UR50D"
    tokenizer_path = "facebook/esm2_t33_650M_UR50D"
    cache_dir = None
    device = "cpu"

    embedder = HuggingFaceEmbedder(
        model_loader,
        tokenizer_loader,
        language_model_path,
        tokenizer_path,
        cache_dir,
        device,
    )
    return embedder


def test_huggingface_embedder(huggingface_embedder):
    protein_sequences = ["MTEITAAMVKELRESTGAGMMDCKNALSETQHEEIAFLKRLME"]
    embeddings = huggingface_embedder.embed(protein_sequences)
    assert isinstance(embeddings, np.ndarray)
    assert embeddings.shape[0] == len(protein_sequences)


def test_sanitize_intervals():
    intervals = [(1, 5), (3, 7), (8, 10)]
    sanitized = sanitize_intervals(intervals)
    assert sanitized == [(1, 7), (8, 10)]


def test_round_up():
    number = 3.14
    rounded = round_up(number)
    assert rounded == 4


def test_sanitize_intervals_with_padding():
    intervals = [(1, 3), (6, 8)]
    padded_intervals = sanitize_intervals_with_padding(
        intervals, pad_value=5, max_value=20
    )
    assert padded_intervals == [(0, 4), (5, 9)]


def test_selection_generator():
    pool_of_sequences = [
        {"sequence": "A", "score": 0.9},
        {"sequence": "B", "score": 0.8},
        {"sequence": "C", "score": 0.95},
        {"sequence": "D", "score": 0.7},
    ]
    generator = SelectionGenerator()
    selected = generator.selection(pool_of_sequences, k=0.5)
    assert len(selected) == 2
    assert selected[0]["sequence"] == "C"


def test_crossover_generator():
    generator = CrossoverGenerator(threshold_probability=0.5)
    seq_a = "AAAAAAAA"
    seq_b = "BBBBBBBB"
    offspring_a, offspring_b = generator.uniform_crossover(seq_a, seq_b)
    assert len(offspring_a) == len(seq_a)
    assert all(c in ["A", "B"] for c in offspring_a)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import Any, Callable, Dict, List, Tuple, Union, Optional

import numpy as np
import torch
import torch.nn as nn
import torch_geometric.data as gd
from rdkit.Chem.rdchem import Mol as RDMol

from gt4sd.frameworks.gflownet.dataloader.dataset import (
    FlatRewards,
    GFlowNetDataset,
    GFlowNetTask,
    RewardScalar,
)
from gt4sd.frameworks.gflownet.ml.models.mxmnet import (
    HAR2EV,
    MXMNet,
    MXMNetConfig,
    mol2graph,
)

PROPERTIES: List[str] = [
    "rA",
    "rB",
    "rC",
    "mu",
    "alpha",
    "homo",
    "lumo",
    "gap",
    "r2",
    "zpve",
    "U0",
    "U",
    "H",
    "G",
    "Cv",
]


class QM9Dataset(GFlowNetDataset):
    """QM9 dataset compatible with gflownet."""

    def __init__(
        self,
        h5_file: str,
        target: str = "gap",
        properties: List[str] = PROPERTIES,
    ) -> None:
        """Initialize QM9 dataset.

        Args:
            h5_file: path to the h5 file containing the dataset.
            target: target property to optimize and build the reward.
            properties: list of properties to use as features.
        """
        super().__init__(
            h5_file=h5_file,
            target=target,
            properties=properties,
        )


def thermometer(
    v: torch.Tensor, n_bins: int = 50, vmin: int = 0, vmax: int = 1
) -> torch.Tensor:
    """Compute a thermometer reward using gap.

    Args:
        v: tensor of values to compute the reward.
        n_bins: number of bins to use.
        vmin: minimum value of the range.
        vmax: maximum value of the range.

    Returns:
        tensor of the reward.
    """
    bins = torch.linspace(vmin, vmax, n_bins)
    gap = bins[1] - bins[0]
    return (v[..., None] - bins.reshape((1,) * v.ndim + (-1,))).clamp(
        0, gap.item()
    ) / gap


# define task
class QM9GapTask(GFlowNetTask):
    """QM9 task compatible with gflownet."""

    def __init__(
        self,
        configuration: Dict[str, Any],
        dataset: GFlowNetDataset,
        reward_model: Optional[nn.Module] = None,
        wrap_model: Optional[Callable[[nn.Module], nn.Module]] = None,
    ):
        """Initialize QM9 task.

        Code adapted from: https://github.com/recursionpharma/gflownet/blob/trunk/src/gflownet/tasks/qm9/qm9.py.

        Args:
            configuration: configuration of the task.
            dataset: dataset to use for the task.
            reward_model: model to use for the reward.
            wrap_model: function to wrap the model.
        """
        super().__init__(
            configuration=configuration,
            dataset=dataset,
            reward_model=reward_model,
            wrap_model=wrap_model,
        )

    def flat_reward_transform(self, _y: Union[float, torch.Tensor]) -> FlatRewards:
        """Transforms a target quantity y (e.g. the LUMO energy in QM9) to a positive reward scalar.

        Args:
            _y: target quantity to transform.

        Returns:
            reward scalar.
        """
        y = np.array(_y)
        y = y.astype("float")
        if self._rtrans == "exp":
            flat_r = np.exp(-(y - self._min) / self._width)
        elif self._rtrans == "unit":
            flat_r = 1 - (y - self._min) / self._width
        elif self._rtrans == "unit+95p":
            # Add constant such that 5% of rewards are > 1
            flat_r = 1 - (y - self._percentile_95) / self._width
        else:
            raise ValueError(self._rtrans)
        return FlatRewards(flat_r)

    def inverse_flat_reward_transform(self, rp):
        """Inverse transform a reward scalar to a target quantity y (e.g. the LUMO energy in QM9).

        Args:
            rp: reward scalar to transform.

        Returns:
            target quantity.
        """
        if self._rtrans == "exp":
            return -np.log(rp) * self._width + self._min
        elif self._rtrans == "unit":
            return (1 - rp) * self._width + self._min
        elif self._rtrans == "unit+95p":
            return (1 - rp + (1 - self._percentile_95)) * self._width + self._min

    def load_task_models(self) -> Dict[str, nn.Module]:
        """Loads the models for the task.

        Returns:
            dictionary of models.
        """
        gap_model = MXMNet(MXMNetConfig(128, 6, 5.0))
        try:
            state_dict = torch.load("/ckpt/mxmnet_gap_model.pt")
            gap_model.load_state_dict(state_dict)
        except FileNotFoundError:
            pass
        gap_model.to(self.device)
        # gap_model = self._wrap_model(gap_model)
        return {"model_task": gap_model}

    def sample_conditional_information(self, n):
        """Sample conditional information for the task.

        Args:
            n: number of samples to sample.

        Returns:
            dictionary of conditional information.
        """
        beta = None
        if self.temperature_sample_dist == "gamma":
            beta = self.rng.gamma(*self.temperature_dist_params, n).astype(np.float32)
        elif self.temperature_sample_dist == "uniform":
            beta = self.rng.uniform(*self.temperature_dist_params, n).astype(np.float32)
        elif self.temperature_sample_dist == "beta":
            beta = self.rng.beta(*self.temperature_dist_params, n).astype(np.float32)
        beta_enc = thermometer(torch.tensor(beta), 32, 0, 32)
        return {"beta": torch.tensor(beta), "encoding": beta_enc}

    def cond_info_to_reward(
        self, cond_info: Dict[str, torch.Tensor], _flat_reward: FlatRewards
    ) -> RewardScalar:

        """Compute the reward for a given conditional information.

        Args:
            cond_info: dictionary of conditional information.
            _flat_reward: flat reward.
        Returns:
            reward scalar.
        """
        if isinstance(_flat_reward, list):
            flat_reward = torch.tensor(_flat_reward)
        return RewardScalar(flat_reward ** cond_info["beta"])

    def compute_flat_rewards(
        self, mols: List[RDMol]
    ) -> Tuple[RewardScalar, torch.Tensor]:
        """Computes the flat rewards for a list of molecules.

        Args:
            mols: list of molecules.

        Returns:
            reward scalar and validity.
        """
        graphs = [mol2graph(i) for i in mols]  # type: ignore[attr-defined]
        is_valid = torch.tensor([i is not None for i in graphs]).bool()
        if not is_valid.any():
            return RewardScalar(torch.zeros((0,))), is_valid

        batch = gd.Batch.from_data_list([i for i in graphs if i is not None])
        batch.to(self.device)

        # sample model
        preds = self.model["model_task"](batch)
        preds = preds.reshape((-1,)).data.cpu() / HAR2EV  # type: ignore[attr-defined]
        preds[preds.isnan()] = 1
        preds = self.flat_reward_transform(preds).clip(1e-4, 2)

        return RewardScalar(preds), is_valid



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""PaccMann training utilities."""

from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from ..core import TrainingPipeline, TrainingPipelineArguments


class PaccMannTrainingPipeline(TrainingPipeline):
    """PyTorch lightining training pipelines."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for PaccMann training.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Raises:
            NotImplementedError: the generic trainer does not implement the pipeline.
        """
        raise NotImplementedError


@dataclass
class PaccMannTrainingArguments(TrainingPipelineArguments):
    """Arguments related to PaccMann trainer."""

    __name__ = "training_args"

    model_path: str = field(
        metadata={"help": "Path where the model artifacts are stored."}
    )
    training_name: str = field(metadata={"help": "Name used to identify the training."})
    checkpoint_path: Optional[str] = field(
        default=None,
        metadata={
            "help": "Path to model checkpoint for weights initialization. Leave None if you want to train a model from scratch"
        },
    )
    epochs: int = field(default=50, metadata={"help": "Number of epochs."})
    batch_size: int = field(default=256, metadata={"help": "Size of the batch."})
    learning_rate: float = field(
        default=0.0005, metadata={"help": "Learning rate used in training."}
    )
    optimizer: str = field(
        default="adam", metadata={"help": "Optimizer used during training."}
    )
    log_interval: int = field(
        default=100, metadata={"help": "Number of steps between log intervals."}
    )
    save_interval: int = field(
        default=1000, metadata={"help": "Number of steps between model save intervals."}
    )
    eval_interval: int = field(
        default=500, metadata={"help": "Number of steps between evaluation intervals."}
    )


@dataclass
class PaccMannDataArguments(TrainingPipelineArguments):
    """Arguments related to PaccMann data loading."""

    __name__ = "dataset_args"

    train_smiles_filepath: str = field(
        metadata={"help": "Training file containing SMILES in .smi format."}
    )
    test_smiles_filepath: str = field(
        metadata={"help": "Testing file containing SMILES in .smi format."}
    )
    smiles_language_filepath: str = field(
        default="none", metadata={"help": "Optional SMILES language file."}
    )
    add_start_stop_token: bool = field(
        default=True, metadata={"help": "Whether start and stop token should be added."}
    )
    selfies: bool = field(
        default=True, metadata={"help": "Whether SELFIES representations are used."}
    )
    num_workers: int = field(
        default=0, metadata={"help": "Number of workers used in data loading."}
    )
    pin_memory: bool = field(
        default=False, metadata={"help": "Whether memory in the data loader is pinned."}
    )
    augment_smiles: bool = field(
        default=False, metadata={"help": "Whether SMILES augumentation is used."}
    )
    canonical: bool = field(
        default=False, metadata={"help": "Whether SMILES canonicalization is used."}
    )
    kekulize: bool = field(
        default=False, metadata={"help": "Whether SMILES kekulization is used."}
    )
    all_bonds_explicit: bool = field(
        default=False, metadata={"help": "Whether all bonds are explicit."}
    )
    all_hs_explicit: bool = field(
        default=False, metadata={"help": "Whether all hydrogens are explicit."}
    )
    remove_bonddir: bool = field(
        default=False, metadata={"help": "Remove bond directionality."}
    )
    remove_chirality: bool = field(
        default=False, metadata={"help": "Remove chirality."}
    )


@dataclass
class PaccMannSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to PaccMann trainer."""

    __name__ = "saving_args"

    model_path: str = field(
        metadata={"help": "Path where the model artifacts are stored."}
    )
    training_name: str = field(metadata={"help": "Name used to identify the training."})



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import List, Tuple

import networkx as nx
import numpy as np
import rdkit.Chem as Chem
import torch
import torch_geometric.data as gd
from rdkit.Chem import Mol
from rdkit.Chem.rdchem import BondType, ChiralType

from ..envs.graph_building_env import (
    Graph,
    GraphAction,
    GraphActionType,
    GraphBuildingEnvContext,
)


class MolBuildingEnvContext(GraphBuildingEnvContext):
    """A context environment for building molecular graphs."""

    def __init__(
        self,
        atoms: List[str] = ["H", "C", "N", "O", "F"],
        num_cond_dim: int = 32,
        device="cpu",
    ) -> None:
        """Initialize a generic context environment for molecules.

        A specification of what is being generated for a GraphBuildingEnv.
        This context specifies how to create molecules atom-by-atom (and attribute-by-attribute).

        Code adapted from: https://github.com/recursionpharma/gflownet/tree/trunk/src/gflownet/envs

        Args:
            atoms: basic building blocks. In principle we can infer this information from the dataset.
            num_cond_dim: number of conditional dimensions.s
            device: device to use (cpu, cuda).
        """

        # idx 0 has to coincide with the default value
        self.atom_attr_values = {
            "v": atoms,
            "chi": [
                ChiralType.CHI_UNSPECIFIED,
                ChiralType.CHI_TETRAHEDRAL_CW,
                ChiralType.CHI_TETRAHEDRAL_CCW,
            ],
            "charge": [0, 1, -1],
            "expl_H": list(range(4)),  # TODO: check what is the actual range of this
            "no_impl": [False, True],
        }
        self.atom_attr_defaults = {
            k: self.atom_attr_values[k][0] for k in self.atom_attr_values
        }
        # The size of the input vector for each atom
        self.atom_attr_size = sum(len(i) for i in self.atom_attr_values.values())
        self.atom_attrs = sorted(self.atom_attr_values.keys())
        # The beginning position within the input vector of each attribute
        self.atom_attr_slice = [0] + list(
            np.cumsum([len(self.atom_attr_values[i]) for i in self.atom_attrs])
        )
        # The beginning position within the logit vector of each attribute
        self.atom_attr_logit_slice = {
            k: s
            for k, s in zip(
                self.atom_attrs,
                [0]
                + list(
                    np.cumsum(
                        [len(self.atom_attr_values[i]) - 1 for i in self.atom_attrs]
                    )
                ),
            )
        }
        # The attribute and value each logit dimension maps back to
        self.atom_attr_logit_map = [
            (k, v)
            for k in self.atom_attrs
            if k != "v"
            # index 0 is skipped because it is the default value
            for v in self.atom_attr_values[k][1:]
        ]

        self.bond_attr_values = {
            "type": [
                BondType.SINGLE,
                BondType.DOUBLE,
                BondType.TRIPLE,
                BondType.AROMATIC,
            ],
        }
        self.bond_attr_defaults = {
            k: self.bond_attr_values[k][0] for k in self.bond_attr_values
        }
        self.bond_attr_size = sum(len(i) for i in self.bond_attr_values.values())
        self.bond_attrs = sorted(self.bond_attr_values.keys())
        self.bond_attr_slice = [0] + list(
            np.cumsum([len(self.bond_attr_values[i]) for i in self.bond_attrs])
        )
        self.bond_attr_logit_slice = {
            k: s
            for k, s in zip(
                self.bond_attrs,
                [0]
                + list(
                    np.cumsum(
                        [len(self.bond_attr_values[i]) - 1 for i in self.bond_attrs]
                    )
                ),
            )
        }
        self.bond_attr_logit_map = [
            (k, v) for k in self.bond_attrs for v in self.bond_attr_values[k][1:]
        ]

        # These values are used by Models to know how many inputs/logits to produce
        self.num_new_node_values = len(atoms)
        self.num_node_attr_logits = len(self.atom_attr_logit_map)
        self.num_node_dim = self.atom_attr_size + 1
        self.num_edge_attr_logits = len(self.bond_attr_logit_map)
        self.num_edge_dim = self.bond_attr_size
        self.num_cond_dim = num_cond_dim

        # Order in which models have to output logits
        self.action_type_order = [
            GraphActionType.Stop,
            GraphActionType.AddNode,
            GraphActionType.SetNodeAttr,
            GraphActionType.AddEdge,
            GraphActionType.SetEdgeAttr,
        ]
        self.device = device

    def aidx_to_graph_action(
        self, g: gd.Data, action_idx: Tuple[int, int, int]
    ) -> GraphAction:
        """Translate an action index (e.g. from a GraphActionCategorical) to a GraphAction.

        Args:
            g: The graph to act on.
            action_idx: The action index.

        Raises:
            ValueError: If the action index is invalid.

        Returns:
            The action corresponding to the action index.
        """

        act_type, act_row, act_col = [int(i) for i in action_idx]
        t = self.action_type_order[act_type]

        if t is GraphActionType.Stop:
            return GraphAction(t)
        elif t is GraphActionType.AddNode:
            return GraphAction(
                t, source=act_row, value=self.atom_attr_values["v"][act_col]
            )
        elif t is GraphActionType.SetNodeAttr:
            attr, val = self.atom_attr_logit_map[act_col]
            return GraphAction(t, source=act_row, attr=attr, value=val)
        elif t is GraphActionType.AddEdge:
            a, b = g.non_edge_index[:, act_row]
            return GraphAction(t, source=a.item(), target=b.item())
        # Edges are duplicated to get undirected GNN, deduplicated for logits
        elif t is GraphActionType.SetEdgeAttr:
            a, b = g.edge_index[:, act_row * 2]
            attr, val = self.bond_attr_logit_map[act_col]
            return GraphAction(
                t, source=a.item(), target=b.item(), attr=attr, value=val
            )
        else:
            raise ValueError(f"Unknown action type: {t}")

    def graph_action_to_aidx(
        self, g: gd.Data, action: GraphAction
    ) -> Tuple[int, int, int]:
        """Translate a GraphAction to an index tuple.

        Args:
            g: The graph to act on.
            action: The action to translate.

        Returns:
            The index corresponding to the action.
        """

        if action.action is GraphActionType.Stop:
            row = 0
            col = 0
        elif action.action is GraphActionType.AddNode:
            row = action.source  # type: ignore
            col = self.atom_attr_values["v"].index(action.value)
        elif action.action is GraphActionType.SetNodeAttr:
            row = action.source  # type: ignore
            # - 1 because the default is index 0
            col = (
                self.atom_attr_values[action.attr].index(action.value)  # type: ignore
                - 1
                + self.atom_attr_logit_slice[action.attr]  # type: ignore
            )
        elif action.action is GraphActionType.AddEdge:
            # Here we have to retrieve the index in non_edge_index of an edge (s,t)
            # that's also possibly in the reverse order (t,s).
            # That's definitely not too efficient, can we do better?
            row = (
                (
                    g.non_edge_index.T == torch.tensor([(action.source, action.target)])
                ).prod(1)
                + (
                    g.non_edge_index.T == torch.tensor([(action.target, action.source)])
                ).prod(1)
            ).argmax()
            col = 0
        elif action.action is GraphActionType.SetEdgeAttr:
            # Here the edges are duplicated, both (i,j) and (j,i) are in edge_index
            # so no need for a double check.
            # row = ((g.edge_index.T == torch.tensor([(action.source, action.target)])).prod(1) +
            #       (g.edge_index.T == torch.tensor([(action.target, action.source)])).prod(1)).argmax()
            row = (
                (g.edge_index.T == torch.tensor([(action.source, action.target)]))
                .prod(1)
                .argmax()
            )
            # Because edges are duplicated but logits aren't, divide by two
            row = row.div(2, rounding_mode="floor")  # type: ignore
            col = (
                self.bond_attr_values[action.attr].index(action.value)  # type: ignore
                - 1
                + self.bond_attr_logit_slice[action.attr]  # type: ignore
            )
        type_idx = self.action_type_order.index(action.action)
        row = int(row)
        col = int(col)
        return (type_idx, row, col)

    def graph_to_data(self, g: Graph) -> gd.Data:
        """Convert a networkx Graph to a torch geometric Data instance.

        Args:
            g: Networkx Graph to convert.

        Returns:
            torch geometric Data instance.
        """

        x = torch.zeros((max(1, len(g.nodes)), self.num_node_dim))
        x[0, -1] = len(g.nodes) == 0

        for i, n in enumerate(g.nodes):
            ad = g.nodes[n]
            for k, sl in zip(self.atom_attrs, self.atom_attr_slice):
                idx = self.atom_attr_values[k].index(ad[k]) if k in ad else 0
                x[i, sl + idx] = 1
        edge_attr = torch.zeros((len(g.edges) * 2, self.num_edge_dim))

        for i, e in enumerate(g.edges):
            ad = g.edges[e]
            for k, sl in zip(self.bond_attrs, self.bond_attr_slice):
                idx = self.bond_attr_values[k].index(ad[k]) if k in ad else 0
                edge_attr[i * 2, sl + idx] = 1
                edge_attr[i * 2 + 1, sl + idx] = 1

        edge_index = torch.tensor(
            [e for i, j in g.edges for e in [(i, j), (j, i)]], dtype=torch.long
        )
        edge_index = edge_index.reshape((-1, 2)).T
        gc = nx.complement(g)

        non_edge_index = torch.tensor([i for i in gc.edges], dtype=torch.long)
        if len(non_edge_index.shape) == 2:
            non_edge_index = non_edge_index.T
        non_edge_index = non_edge_index.reshape((2, -1))
        return gd.Data(x, edge_index, edge_attr, non_edge_index=non_edge_index)

    def collate(self, graphs: List[gd.Data]):
        """Batch Data instances.

        Args:
            graphs: List of Data instances.

        Returns:
            Batch of Data instances.
        """

        return gd.Batch.from_data_list(
            graphs, follow_batch=["edge_index", "non_edge_index"]
        )

    def mol_to_graph(self, mol: Mol) -> Graph:
        """Convert an RDMol to a Graph.

        Args:
            mol: RDKit molecule format.

        Returns:
            Graph format.
        """

        g = Graph()
        # Only set an attribute tag if it is not the default attribute
        for a in mol.GetAtoms():
            attrs = {
                "chi": a.GetChiralTag(),
                "charge": a.GetFormalCharge(),
                "expl_H": a.GetNumExplicitHs(),
                "no_impl": a.GetNoImplicit(),
            }
            g.add_node(
                a.GetIdx(),
                v=a.GetSymbol(),
                **{
                    attr: val
                    for attr, val in attrs.items()
                    if val != self.atom_attr_defaults[attr]
                },
            )
        for b in mol.GetBonds():
            attrs = {"type": b.GetBondType()}
            g.add_edge(
                b.GetBeginAtomIdx(),
                b.GetEndAtomIdx(),
                **{
                    attr: val
                    for attr, val in attrs.items()
                    if val != self.bond_attr_defaults[attr]
                },
            )
        return g

    def graph_to_mol(self, g: Graph) -> Mol:
        """Convert a Graph to an RDKit molecule.

        Args:
            g: Graph format.

        Returns:
            RDKit molecule format.
        """

        mp = Chem.RWMol()
        mp.BeginBatchEdit()
        for i in range(len(g.nodes)):
            d = g.nodes[i]
            a = Chem.Atom(d["v"])
            if "chi" in d:
                a.SetChiralTag(d["chi"])
            if "charge" in d:
                a.SetFormalCharge(d["charge"])
            if "expl_H" in d:
                a.SetNumExplicitHs(d["expl_H"])
            if "no_impl" in d:
                a.SetNoImplicit(d["no_impl"])
            mp.AddAtom(a)
        for e in g.edges:
            d = g.edges[e]
            mp.AddBond(e[0], e[1], d.get("type", BondType.SINGLE))
        mp.CommitBatchEdit()
        Chem.SanitizeMol(mp)
        return mp

    def is_sane(self, g: Graph) -> bool:
        """Check if a graph is sane.

        Args:
            g: Graph format.

        Returns:
            True if sane, False otherwise.
        """
        try:
            mol = self.graph_to_mol(g)
            assert Chem.MolFromSmiles(Chem.MolToSmiles(mol)) is not None
        except Exception:
            return False
        if mol is None:
            return False
        return True



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Tests for AlgorithmConfiguration."""

import os
import shutil
from typing import ClassVar

import pytest

from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.configuration import GT4SDConfiguration

gt4sd_configuration_instance = GT4SDConfiguration.get_instance()


@pytest.fixture()
def development_version_path():
    # setup
    path = os.path.join(
        gt4sd_configuration_instance.gt4sd_local_cache_path,
        gt4sd_configuration_instance.gt4sd_local_cache_path_algorithms,
        "dummy",
        "algorithm",
        "config",
        "development",
    )
    os.makedirs(path, exist_ok=False)
    # test
    yield path
    # teardown
    shutil.rmtree(
        os.path.join(
            gt4sd_configuration_instance.gt4sd_local_cache_path,
            gt4sd_configuration_instance.gt4sd_local_cache_path_algorithms,
            "dummy",
        )
    )


def test_list_versions_local_only(development_version_path):
    class Config(AlgorithmConfiguration):
        algorithm_type: ClassVar[str] = "dummy"
        domain: ClassVar[str] = ""
        algorithm_name: ClassVar[str] = "algorithm"
        algorithm_application: ClassVar[str] = "config"
        algorithm_version: str = "development"

    assert "development" in Config.list_versions()



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Types, classes, validation, etc. for the material domain."""

from enum import Enum
from typing import Any, List, NewType, Tuple, Union

import numpy as np
import pandas as pd
from rdkit import Chem
from rdkit.Chem import Mol

# TODO: setting to str directly requires no wrapping, so wrong strings could be passed
Protein = str  # NewType('Protein', str)
SMILES = str  # NewType('SMILES', str)
SELFIES = str  # NewType('SELFIES', str)
Copolymer = str  # NewType('Copolymer', str)
SmallMolecule = Union[SMILES, Mol]
MacroMolecule = Union[Protein, Mol]
Omics = Union[np.ndarray, pd.Series]
PAG = SMILES
Molecule = Union[SmallMolecule, MacroMolecule]
Sequence = str


class MoleculeFormat(str, Enum):
    selfies = "SELFIES"
    smiles = "SMILES"
    copolymer = "Copolymer"


def validate_smiles(
    smiles_list: List[SMILES],
) -> Tuple[List[Chem.rdchem.Mol], List[int]]:
    """Validate molecules.

    Args:
        smiles_list: list of SMILES representing molecules.

    Returns:
        a tuple containing RDKit molecules and valid indexes.
    """
    # generate molecules from SMILES
    molecules = [
        Chem.MolFromSmiles(a_smiles, sanitize=True) for a_smiles in smiles_list
    ]
    # valid ids
    valid_ids = [
        index
        for index, molecule in enumerate(molecules)
        if molecule is not None and molecule != ""
    ]
    return molecules, valid_ids  # type: ignore


def validate_selfies(
    selfies_list: List[SELFIES],
) -> Tuple[List[SELFIES], List[int]]:
    """Validate molecules.

    Args:
        selfies_list: list of SELFIES representing molecules.

    Returns:
        a tuple containing RDKit molecules and valid indexes.
    """
    # selfies is not valid if it contains -1 so return no valid ids
    valid_ids = [i for i, s in enumerate(selfies_list) if s != -1]
    selfies = [s for i, s in enumerate(selfies_list) if i in valid_ids]
    return selfies, valid_ids


def validate_copolymer(
    copolymers_list: List[Copolymer],
) -> Tuple[List[Copolymer], List[int]]:
    """Validate copolymers.

    Args:
        copolymers_list: list of Copolymer representing molecules.

    Returns:
        a tuple containing RDKit molecules and valid indexes.
    """
    # TODO implement actual validation

    # Remove duplicates
    copolymers, idxs = [], []
    for i, copolymer in enumerate(copolymers_list):
        if copolymer not in copolymers:
            copolymers.append(copolymer)
            idxs.append(i)
    return copolymers, idxs


MOLECULE_FORMAT_VALIDATOR_FACTORY = {
    MoleculeFormat.selfies: validate_selfies,
    MoleculeFormat.smiles: validate_smiles,
    MoleculeFormat.copolymer: validate_copolymer,
}


def validate_molecules(
    pattern_list: List[str],
    input_type: str,
) -> Tuple[List[Any], List[int]]:
    """Validate molecules.

    Args:
        pattern_list: list of patterns representing molecules.
        input_type: type of patter (SELFIES, SMILES OR Copolymer).

    Returns:
        a tuple containing RDKit molecules and valid indexes.
    """
    return MOLECULE_FORMAT_VALIDATOR_FACTORY[input_type](pattern_list)  # type: ignore


Bounds = Tuple[int, int]  # NewType('Bounds', Tuple[int, int])

PhotoacidityCondition = NewType("PhotoacidityCondition", Bounds)

ConditionPAG = Union[PhotoacidityCondition]
